from bioel.models.arboel.biencoder.model.BiEncoderLightningModule import LitArboel
from bioel.models.arboel.crossencoder.model.CrossEncoderLightningModule import (
    LitCrossEncoder,
)
from bioel.models.arboel.biencoder.model.common.params import BlinkParser
from bioel.models.krissbert.model.model import Krissbert
from bioel.models.biogenel.LightningModule import BioGenElLightningModule
from bioel.models.sapbert.model.metric_learning import Sap_Metric_Learning
from bioel.models.sapbert.model.model_wrapper import Model_Wrapper
from bioel.models.sapbert.train import parse_args

import json
import importlib
import argparse
import lightning as L
import os
import torch


class BioEL_Model:
    def __init__(self, model, name, train_script_path, evaluate_script_path, params):
        self.name = name  # name of the model
        self.model = model  # model object
        self.train_script_path = train_script_path  # path to the training script
        self.evaluate_script_path = (
            evaluate_script_path  # path to the evaluation script
        )
        self.params = params  # model arguments for training and inferencing

    @classmethod
    def load_arboel_biencoder(cls, name, params_file, checkpoint_path=None):
        # Check if params_file is a path or a dictionary
        if isinstance(params_file, str) and os.path.isfile(params_file):
            with open(params_file, "r") as f:
                json_params = json.load(f)
        elif isinstance(params_file, dict):
            json_params = params_file
        else:
            raise TypeError("params_file must be a valid file path or a dictionary")

        parser = BlinkParser(add_model_args=True)
        parser.add_training_args()
        parser.add_eval_args()

        # Set default values and remove required flag for params in JSON
        for action in parser._actions:
            if action.dest in json_params:
                parser.set_defaults(**{action.dest: json_params[action.dest]})
                action.required = False

        # Parse arguments
        args = parser.parse_args()
        params = vars(args)
        params.update(json_params)

        if checkpoint_path:
            model = LitArboel.load_from_checkpoint(
                checkpoint_path=checkpoint_path, params=params
            )
        else:
            model = LitArboel(params=params)

        train_script_path = f"bioel/models/arboel/biencoder/train_biencoder.py"
        evaluate_script_path = f"bioel/models/arboel/biencoder/evaluate_biencoder.py"
        return cls(
            model,
            name,
            train_script_path,
            evaluate_script_path,
            params,
        )

    @classmethod
    def load_arboel_crossencoder(cls, name, params_file, checkpoint_path=None):
        # Check if params_file is a path or a dictionary
        if isinstance(params_file, str) and os.path.isfile(params_file):
            with open(params_file, "r") as f:
                json_params = json.load(f)
        elif isinstance(params_file, dict):
            json_params = params_file
        else:
            raise TypeError("params_file must be a valid file path or a dictionary")

        parser = BlinkParser(add_model_args=True)
        parser.add_training_args()
        parser.add_eval_args()

        # Set default values and remove required flag for params in JSON
        for action in parser._actions:
            if action.dest in json_params:
                parser.set_defaults(**{action.dest: json_params[action.dest]})
                action.required = False

        # Parse arguments
        args = parser.parse_args()
        params = vars(args)
        params.update(json_params)

        if checkpoint_path:
            model = LitCrossEncoder.load_from_checkpoint(
                checkpoint_path=checkpoint_path, params=params
            )
        else:
            model = LitCrossEncoder(params=params)

        train_script_path = f"bioel/models/arboel/crossencoder/train_crossencoder.py"
        evaluate_script_path = (
            f"bioel/models/arboel/crossencoder/evaluate_crossencoder.py"
        )
        return cls(
            model,
            name,
            train_script_path,
            evaluate_script_path,
            params,
        )

    @classmethod
    def load_sapbert(cls, name, params_file=None, checkpoint_path=None):
        if isinstance(params_file, str) and os.path.isfile(params_file):
            with open(params_file, "r") as f:
                json_params = json.load(f)
        elif isinstance(params_file, dict):
            json_params = params_file
        else:
            raise TypeError("params_file must be a valid file path or a dictionary")

        params = parse_args(json_params)

        model = Model_Wrapper()  # model wrapper

        if checkpoint_path:
            # Load the pretrained model from the specified path
            model.load_pretrained_model(checkpoint_path, use_cuda=params["use_cuda"])

        train_script_path = f"bioel/models/sapbert/train.py"
        evaluate_script_path = f"bioel/models/sapbert/evaluate.py"

        return cls(
            model,  # model wrapper
            name,
            train_script_path,
            evaluate_script_path,
            params,
        )

    @classmethod
    def load_krissbert(cls, name, params_file, checkpoint_path=None):
        if isinstance(params_file, str) and os.path.isfile(params_file):
            with open(params_file, "r") as f:
                params = json.load(f)
        elif isinstance(params_file, dict):
            params = params_file
        else:
            raise TypeError("params_file must be a valid file path or a dictionary")

        if checkpoint_path:
            model = Krissbert(checkpoint_path)
        else:
            model = Krissbert(params["model_name_or_path"])

        train_script_path = f"bioel/models/krissbert/train.py"
        evaluate_script_path = f"bioel/models/krissbert/evaluate.py"
        return cls(
            model,
            name,
            train_script_path,
            evaluate_script_path,
            params,
        )

    @classmethod
    def load_scispacy(cls, name, params_file):
        # No model object, just a string identifier
        model = "scispacy"
        train_script_path = None  # ScispaCy does not require training
        evaluate_script_path = "bioel/models/scispacy/evaluate.py"

        return cls(model, name, train_script_path, evaluate_script_path, params_file)

    @classmethod
    def load_biobart(cls, name, params_file):
        params = Config(params_file)
        train_script_path = f"bioel/models/biogenel/train_biogenel_biobart.py"
        evaluate_script_path = f"bioel/models/biogenel/eval_biogenel_biobart.py"
        model = BioGenElLightningModule(config=params)
        return cls(
            model,
            name,
            train_script_path,
            evaluate_script_path,
            params,
        )

    @classmethod
    def load_biogenel(cls, name, params_file):
        params = Config(params_file)
        train_script_path = f"bioel/models/biogenel/train_biogenel_biobart.py"
        evaluate_script_path = f"bioel/models/biogenel/eval_biogenel_biobart.py"
        model = BioGenElLightningModule(config=params)
        return cls(
            model,
            name,
            train_script_path,
            evaluate_script_path,
            params,
        )

    def training(self):
        if self.name.lower() == "scispacy":
            print('ScispaCy "training" is done directly in the inference.')
        else:
            module_name = self.train_script_path.replace("/", ".").rsplit(".", 1)[0]
            train_module = importlib.import_module(module_name)
            train_module.train_model(params=self.params, model=self.model)

    def inference(self):
        module_name = self.evaluate_script_path.replace("/", ".").rsplit(".", 1)[0]
        evaluate_module = importlib.import_module(module_name)
        evaluate_module.evaluate_model(self.params, self.model)


class Config:
    def __init__(self, json_file=None):
        if json_file:
            self.load_from_json(json_file)

    def load_from_json(self, json_file):
        with open(json_file, "r") as f:
            json_params = json.load(f)
            for key, value in json_params.items():
                setattr(self, key, value)

    def __repr__(self):
        return f"Config({self.__dict__})"


if __name__ == "__main__":
    # print("Start work on arboel_biencoder")
    # arboel_biencoder_model = BioEL_Model.load_arboel_biencoder(
    #     name="arboel_biencoder",
    #     params_file="/home2/cye73/data_test2/arboel/medmentions_st21pv/params_biencoder.json",
    # )
    # arboel_biencoder_model.training()
    # arboel_biencoder_model.inference()

    # print("Start work on arboel_crossencoder")
    # arboel_crossencoder_model = BioEL_Model.load_arboel_crossencoder(
    #     name="arboel_bcrossencoder",
    #     params_file="/home2/cye73/data_test2/arboel/medmentions_st21pv/params_crossencoder.json",
    #     # checkpoint_path="/home2/cye73/results2/arboel/medmentions_st21pv/crossencoder_2024-09-25_13-22-32-epoch=1-Accuracy=0.89.ckpt",
    # )
    # arboel_crossencoder_model.training()
    # arboel_crossencoder_model.inference()

    # print("Start work on biobart")
    # biobart_model = BioEL_Model.load_biobart(
    #     name="biobart",
    #     params_file="/home2/cye73/data_test2/biogenel/ncbi_config.json",
    # )
    # biobart_model.training()
    # biobart_model.inference()

    # # krissbert = BioEL_Model.load_krissbert(
    # #     name="krissbert",
    # #     params_file="/home2/cye73/data_test2/krissbert/ncbi_disease/params.json",
    # # )
    # # krissbert.training()
    # # krissbert.inference()

    # scispacy_params = {
    #     "dataset": "ncbi_disease",
    #     "load_function": "load_medic",
    #     "ontology_dict": {
    #         "name": "medic",
    #         "filepath": "/mitchell/entity-linking/kbs/medic.tsv",
    #     },
    #     "k": 10,
    #     "path_to_save": "/home2/cye73/data_test2/scispacy/kb_paths_scispacy/ncbi_disease",
    #     "output_path": "/home2/cye73/results2/scispacy/ncbi_disease_output.json",
    #     "equivalant_cuis": True,
    #     "path_to_abbrev": "/home2/cye73/data_test2/abbreviations.json",
    # }
    # scispacy = BioEL_Model.load_scispacy(name="scispacy", params_file=scispacy_params)
    # scispacy.training()
    # scispacy.inference()

    # dataset = "nlm_gene"
    dataset = "nlmchem"
    print("Start work on sapbert :", dataset)
    sapbert_model = BioEL_Model.load_sapbert(
        name="sapbert",
        params_file=f"/home2/cye73/data_test2/sapbert/{dataset}/params.json",
        checkpoint_path=f"/home2/cye73/data_test2/sapbert/{dataset}/finetuned_models_nice/eager-voice-55",
    )
    # sapbert_model.training()
    sapbert_model.inference()



import os
import sys
import numpy as np
from tqdm import tqdm
import pickle
import json
import argparse
import pickle
import pandas as pd
from transformers import BartTokenizer, AutoTokenizer
from bioel.models.biogenel.trie import Trie

sys.setrecursionlimit(8000)


def create_trie(data_dir: str, use_biobart_tokenizer=False):
    """
    Function that creates the Trie given a repository director containing the target knowledge base (ontology) and a tokenizer
    """
    if use_biobart_tokenizer:
        tokenizer = AutoTokenizer.from_pretrained("GanjinZero/biobart-v2-large")
    else:
        tokenizer = BartTokenizer.from_pretrained("facebook/bart-large")

    with open(f"{data_dir}/target_kb.json", "r") as f:
        cui2str = json.load(f)

    entities = []
    for cui in cui2str:
        entities += cui2str[cui]

    trie = Trie(
        [16] + list(tokenizer(" " + entity.lower())["input_ids"][1:])
        for entity in tqdm(entities)
    ).trie_dict

    if use_biobart_tokenizer:
        with open(f"{data_dir}/biobart_trie.pkl", "wb") as w_f:
            pickle.dump(trie, w_f)
    else:
        with open(f"{data_dir}/trie.pkl", "wb") as w_f:
            pickle.dump(trie, w_f)

    print("finish running!")



# Copyright (c) Facebook, Inc. and its affiliates.
# Copyright (c) 2021 Dhruv Agarwal and authors of arboEL.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

import os
import random
import time
import pickle
import math
import json
import copy

import numpy as np
import torch
from torch.utils.data import (DataLoader, RandomSampler, SequentialSampler)
from pytorch_transformers.optimization import WarmupLinearSchedule
from tqdm import tqdm, trange

from bioel.models.arboel.model.special_partition.special_partition import (
    cluster_linking_partition,
)
from bioel.models.arboel.crossencoder.original.train_cross import read_dataset, write_to_file, save_model
import bioel.models.arboel.data.data_process as data_process
import bioel.models.arboel.model.eval_cluster_linking as eval_cluster_linking

from bioel.models.arboel.model.biencoder import BiEncoderRanker
from bioel.models.arboel.crossencoder.crossencoder import CrossEncoderRanker
from bioel.models.arboel.model.common.optimizer import get_bert_optimizer
from bioel.models.arboel.model.common.params import BlinkParser

from bioel.logger import setup_logger


from IPython import embed

SCORING_BATCH_SIZE = 64

OPTIM_SCHED_FNAME = "optim_sched.pth"

def concat_for_crossencoder(context_inputs, candidate_inputs, max_seq_length):
    """
    Parameters
    ----------
    context_inputs: N x D
    candidate_inputs: N x k x D
    max_seq_length: int

    Returns
    -------
    concat_input: N x k x 2D-1
    """
    new_input = []
    context_inputs = context_inputs.tolist()
    candidate_inputs = candidate_inputs.tolist()
    for i in range(len(context_inputs)):
        cur_input = context_inputs[i]
        cur_candidates = candidate_inputs[i]
        mod_input = []
        for j in range(len(cur_candidates)):
            # remove [CLS] token from candidate
            sample = cur_input + cur_candidates[j][1:]
            sample = sample[:max_seq_length]
            mod_input.append(sample)
        new_input.append(mod_input)
    return torch.LongTensor(new_input)


def score_in_batches(cross_reranker, max_context_length, cross_inputs, is_context_encoder, silent=False, scoring_batch_size=None):
    batch_size = scoring_batch_size if scoring_batch_size is not None else SCORING_BATCH_SIZE
    n_inputs, scores, reshaped = len(cross_inputs), None, False
    _iter = [cross_inputs]
    if n_inputs > 1:
        sampler = SequentialSampler(cross_inputs)
        dataloader = DataLoader(
            cross_inputs, sampler=sampler, batch_size=batch_size
        )
        _iter = dataloader if silent else tqdm(dataloader, desc=f"Scoring in batches (size={batch_size})")
    for batch in _iter:
        batch_shape = batch.size()
        if batch_shape[0] < cross_reranker.n_gpu:
            reshaped = True
            # Repeat to get length of 0th dimension >= n_gpu
            # NOTE: Needed to avoid CUDA error with data parallel
            repeat_shape = [math.ceil(cross_reranker.n_gpu / float(batch_shape[0])), *[1]*(len(batch_shape) - 1)]
            batch = batch.repeat(repeat_shape)[:cross_reranker.n_gpu]
        batch_scores = cross_reranker.score_candidate(batch.cuda(),
                                                      max_context_length,
                                                      is_context_encoder=is_context_encoder)
        if reshaped:
            reshaped = False
            batch_scores = batch_scores[:batch_shape[0]]
        scores = batch_scores if scores is None else torch.cat((scores, batch_scores), dim=0)
    return scores


def evaluate(cross_reranker,
             max_context_length,
             entity_dictionary,
             valid_processed_data,
             biencoder_valid_idxs,
             bi_valid_nn_count,
             valid_men_inputs,
             valid_ent_inputs,
             logger,
             max_k=8,
             k_biencoder=64):
    """
        - get 64 nearest bi-encoder mentions and entities
        - score using cross-encoder
        - create k-nn graphs
        - prune & predict
    """
    n_entities = len(entity_dictionary)
    n_mentions = len(valid_processed_data)

    bi_men_idxs = biencoder_valid_idxs['men_nns'][:, :k_biencoder]
    bi_ent_idxs = biencoder_valid_idxs['dict_nns'][:, :k_biencoder]

    joint_graphs = {}
    for k in ([0] + [2 ** i for i in range(int(math.log(max_k, 2)) + 1)]):
        joint_graphs[k] = {
            'rows': np.array([]),
            'cols': np.array([]),
            'data': np.array([]),
            'shape': (n_entities + n_mentions, n_entities + n_mentions)
        }
        max_k = k  # Stores the number of mention edges to add per query mention

    torch.cuda.empty_cache()
    cross_reranker.model.eval()
    with torch.no_grad():
        logger.info('Eval: Scoring mention-mention edges using cross-encoder...')
        cross_men_scores = score_in_batches(cross_reranker, max_context_length, valid_men_inputs,
                                            is_context_encoder=True)
        for i in range(len(cross_men_scores)):
            # Set scores for all invalid nearest neighbours to -infinity (due to variable NN counts of mentions)
            cross_men_scores[i][bi_valid_nn_count[i]:] = float('-inf')
        cross_men_topk_scores, cross_men_topk_idxs = torch.sort(cross_men_scores, dim=1, descending=True)
        cross_men_topk_idxs = cross_men_topk_idxs.cpu()[:, :max_k]
        cross_men_topk_scores = cross_men_topk_scores.cpu()[:, :max_k]
        logger.info('Eval: Scoring done')

        logger.info('Eval: Scoring mention-entity edges using cross-encoder...')
        cross_ent_scores = score_in_batches(cross_reranker, max_context_length, valid_ent_inputs,
                                            is_context_encoder=False)
        cross_ent_top1_score, cross_ent_top1_idx = torch.sort(cross_ent_scores, dim=1, descending=True)
        cross_ent_top1_idx = cross_ent_top1_idx.cpu()[:, 0]
        cross_ent_top1_score = cross_ent_top1_score.cpu()[:, 0]
        logger.info('Eval: Scoring done')

    bi_recall = 0.
    # TODO: Add cross-encoder recall metrics

    for men_idx in tqdm(range(len(valid_processed_data)), total=len(valid_processed_data), desc="Eval: Building graphs"):
        # Track biencoder recall@<k_biencoder>
        gold_idx = valid_processed_data[men_idx]["label_idxs"][0]
        if gold_idx in bi_ent_idxs[men_idx]:
            bi_recall += 1.
        # Get nearest entity
        m_e_idx = bi_ent_idxs[men_idx, cross_ent_top1_idx[men_idx]]
        m_e_score = cross_ent_top1_score[men_idx]
        if bi_valid_nn_count[men_idx] > 0:
            # Get nearest mentions
            topk_defined_nn_idxs = cross_men_topk_idxs[men_idx][:bi_valid_nn_count[men_idx]]
            m_m_idxs = bi_men_idxs[men_idx, topk_defined_nn_idxs] + n_entities  # Mentions added at an offset of maximum entities
            if len(topk_defined_nn_idxs) == 1:
                m_m_idxs = [m_m_idxs]
            m_m_scores = cross_men_topk_scores[men_idx][:bi_valid_nn_count[men_idx]]
        # Add edges to the graphs
        for k in joint_graphs:
            # Add mention-entity edge
            joint_graphs[k]['rows'] = np.append(
                joint_graphs[k]['rows'], [n_entities + men_idx])  # Mentions added at an offset of maximum entities
            joint_graphs[k]['cols'] = np.append(
                joint_graphs[k]['cols'], m_e_idx)
            joint_graphs[k]['data'] = np.append(
                joint_graphs[k]['data'], m_e_score)
            if k > 0 and bi_valid_nn_count[men_idx] > 0:
                # Add mention-mention edges
                joint_graphs[k]['rows'] = np.append(
                    joint_graphs[k]['rows'], [n_entities + men_idx] * len(m_m_idxs[:k]))
                joint_graphs[k]['cols'] = np.append(
                    joint_graphs[k]['cols'], m_m_idxs[:k])
                joint_graphs[k]['data'] = np.append(
                    joint_graphs[k]['data'], m_m_scores[:k])

    # Partition graphs and analyze clusters for final predictions
    max_eval_acc = {
        'directed': -1.,
        'directed_k': 0,
        'undirected': -1.,
        'undirected_k': 0
    }
    for k in joint_graphs:
        for mode in ['directed', 'undirected']:
            if k == 0 and mode == 'undirected':
                continue  # Since @k=0, both modes are equivalent
            logger.info(f"\nEval: Graph (k={k}, mode={mode}):")
            # Partition graph based on cluster-linking constraints
            partitioned_graph, clusters = eval_cluster_linking.partition_graph(
                joint_graphs[k], n_entities, directed=(mode == 'directed'), return_clusters=True)
            # Infer predictions from clusters
            result = eval_cluster_linking.analyzeClusters(clusters, entity_dictionary, valid_processed_data, k)
            acc = float(result['accuracy'].split(' ')[0])
            if acc > max_eval_acc[mode]:
                max_eval_acc[mode] = acc
                max_eval_acc[f'{mode}_k'] = k
                if k == 0 and mode == 'directed':
                    max_eval_acc['undirected'] = acc
            logger.info(f"Eval: accuracy for graph@(k={k}, mode={mode}): {acc}%")
    # Compute biencoder recall
    bi_recall /= len(valid_processed_data)
    logger.info(f"Eval: Biencoder recall@{k_biencoder} = {bi_recall * 100}%")
    logger.info(f"Eval: Best cross-encoder accuracy = {json.dumps(max_eval_acc)}")
    return max_eval_acc


def get_optimizer(model, params):
    return get_bert_optimizer(
        [model],
        params["type_optimization"],
        params["learning_rate"],
        fp16=params.get("fp16"),
    )

def get_scheduler(params, optimizer, len_train_data, logger):
    batch_size = params["train_batch_size"]
    grad_acc = params["gradient_accumulation_steps"]
    epochs = params["num_train_epochs"]

    num_train_steps = int(len_train_data / batch_size / grad_acc) * epochs
    num_warmup_steps = int(num_train_steps * params["warmup_proportion"])

    scheduler = WarmupLinearSchedule(
        optimizer, warmup_steps=num_warmup_steps, t_total=num_train_steps,
    )
    logger.info(" Num optimization steps = %d" % num_train_steps)
    logger.info(" Num warmup steps = %d", num_warmup_steps)
    return scheduler


def load_optimizer_scheduler(params, logger):
    optim_sched = None
    model_path = params["path_to_model"]
    if model_path is not None:
        model_dir = os.path.dirname(model_path)
        optim_sched_fpath = os.path.join(model_dir, OPTIM_SCHED_FNAME)
        if os.path.isfile(optim_sched_fpath):
            logger.info(f'Loading stored optimizer and scheduler from {optim_sched_fpath}')
            optim_sched = torch.load(optim_sched_fpath)
    return optim_sched


def get_context_doc_ids(split, params):
    samples = read_dataset(split, params["data_path"])
    # Filter unlabeled data
    has_mult_labels = "labels" in samples[0].keys()
    samples = list(
        filter(lambda sample: (len(sample["labels"]) > 0) if has_mult_labels else (sample["label"] is not None),
               samples))
    context_doc_ids = [s['context_doc_id'] for s in samples]
    return context_doc_ids


def filter_by_context_doc_id(mention_idxs, doc_id, doc_id_list, return_numpy=False):
    mask = [doc_id_list[i] == doc_id for i in mention_idxs]
    if isinstance(mention_idxs, list):
        mention_idxs = np.array(mention_idxs)
    mention_idxs = mention_idxs[mask]
    if not return_numpy:
        mention_idxs = list(mention_idxs)
    return mention_idxs, mask


def get_biencoder_nns(bi_reranker, biencoder_indices_path, entity_dictionary, entity_dict_vecs, train_men_vecs,
                      train_processed_data, train_gold_clusters, valid_men_vecs, valid_processed_data,
                      use_types, logger, n_gpu, params, train_context_doc_ids=None, valid_context_doc_ids=None):
    """
    Train:
        Store sorted entities and mention idxs to disk
            <k_dict_nns> nearest non-gold (negative) entities
            <k_men_nns> nearest negative mentions
            min(<k_men_nns>, len(gold_cluster)-1) nearest gold mentions
    Dev:
        Store sorted entities and mention idxs to disk
            <k_dict_nns> nearest entities
            <k_men_nns> nearest mentions
    """
    dict_embeddings = None
    biencoder_train_idxs, biencoder_valid_idxs = None, None
    # Number of nearest neighbors to fetch and persist
    k_dict_nns, k_men_nns = 128, 128
    within_doc = params["within_doc"]

    if not params["only_evaluate"]:
        # Train set
        biencoder_train_idxs_pkl_path = os.path.join(biencoder_indices_path, 'biencoder_train_idxs.pickle')
        if os.path.isfile(biencoder_train_idxs_pkl_path):
            logger.info("Loading stored sorted biencoder train indices...")
            with open(biencoder_train_idxs_pkl_path, 'rb') as read_handle:
                biencoder_train_idxs = pickle.load(read_handle)
        else:
            logger.info('Biencoder: Embedding and indexing training data')
            if use_types:
                dict_embeddings, dict_indexes, dict_idxs_by_type = data_process.embed_and_index(
                    bi_reranker, entity_dict_vecs, encoder_type="candidate", n_gpu=n_gpu, corpus=entity_dictionary,
                    force_exact_search=True, batch_size=params['embed_batch_size'])
                train_men_embeddings, train_men_indexes, men_idxs_by_type = data_process.embed_and_index(
                    bi_reranker, train_men_vecs, encoder_type="context", n_gpu=n_gpu, corpus=train_processed_data,
                    force_exact_search=True, batch_size=params['embed_batch_size'])
            else:
                dict_embeddings, dict_index = data_process.embed_and_index(bi_reranker, entity_dict_vecs,
                    encoder_type="candidate", n_gpu=n_gpu, force_exact_search=True, batch_size=params['embed_batch_size'])
                train_men_embeddings, train_men_index = data_process.embed_and_index(bi_reranker, train_men_vecs,
                    encoder_type="context", n_gpu=n_gpu, force_exact_search=True, batch_size=params['embed_batch_size'])
            logger.info('Biencoder: Embedding and indexing finished')

            logger.info("Biencoder: Finding nearest mentions and entities for each mention...")
            bi_dict_nns = np.zeros((len(train_men_embeddings), k_dict_nns), dtype=int)  # negatives
            bi_men_nns = -1 * np.ones((len(train_men_embeddings), k_men_nns), dtype=int)  # negatives
            bi_men_gold_nns = -1 * np.ones((len(train_men_embeddings), k_men_nns), dtype=int)  # positives
            if not use_types:
                _, bi_dict_nns_np = dict_index.search(train_men_embeddings, k_dict_nns+1)
                _, bi_men_nns_np = train_men_index.search(train_men_embeddings, len(train_men_embeddings))
                for i in range(len(bi_dict_nns_np)):
                    gold_idx = train_processed_data[i]['label_idxs'][0]
                    bi_dict_nns[i] = bi_dict_nns_np[i][bi_dict_nns_np[i] != gold_idx][:k_dict_nns]
                    men_neg_mask = ~np.isin(bi_men_nns_np[i], train_gold_clusters[gold_idx])
                    neg_nns = bi_men_nns_np[i][men_neg_mask]
                    if within_doc:
                        neg_nns, _ = filter_by_context_doc_id(neg_nns, train_context_doc_ids[i], train_context_doc_ids,
                                                           return_numpy=True)
                    bi_men_nns[i][:len(neg_nns[:k_men_nns])] = neg_nns[:k_men_nns]
                    men_gold_mask = (np.isin(bi_men_nns_np[i], train_gold_clusters[gold_idx])) & (bi_men_nns_np[i] != i)
                    gold_nns = bi_men_nns_np[i][men_gold_mask]
                    if within_doc:
                        gold_nns, _ = filter_by_context_doc_id(gold_nns, train_context_doc_ids[i], train_context_doc_ids,
                                                            return_numpy=True)
                    # If len(gold_nns) < k_men_nns then set bi_men_gold_nns[i] = [...gold_nns, -1, -1, ...]
                    bi_men_gold_nns[i][:len(gold_nns[:k_men_nns])] = gold_nns[:k_men_nns]
            else:
                for entity_type in train_men_indexes:
                    men_embeds_by_type = train_men_embeddings[men_idxs_by_type[entity_type]]
                    _, dict_nns_by_type = dict_indexes[entity_type].search(men_embeds_by_type, k_dict_nns+1)
                    _, men_nns_by_type = train_men_indexes[entity_type].search(men_embeds_by_type, len(men_embeds_by_type))
                    dict_nns_idxs = np.array(list(map(lambda x: dict_idxs_by_type[entity_type][x], dict_nns_by_type)))
                    men_nns_idxs = np.array(list(map(lambda x: men_idxs_by_type[entity_type][x], men_nns_by_type)))
                    for i, idx in enumerate(men_idxs_by_type[entity_type]):
                        gold_idx = train_processed_data[idx]['label_idxs'][0]
                        bi_dict_nns[idx] = dict_nns_idxs[i][dict_nns_idxs[i] != gold_idx][:k_dict_nns]
                        men_neg_mask = ~np.isin(men_nns_idxs[i], train_gold_clusters[gold_idx])
                        neg_nns = men_nns_idxs[i][men_neg_mask]
                        if within_doc:
                            neg_nns, _ = filter_by_context_doc_id(neg_nns, train_context_doc_ids[idx],
                                                               train_context_doc_ids,
                                                               return_numpy=True)
                        bi_men_nns[idx][:len(neg_nns[:k_men_nns])] = neg_nns[:k_men_nns]
                        men_gold_mask = (np.isin(men_nns_idxs[i], train_gold_clusters[gold_idx])) & (men_nns_idxs[i] != idx)
                        gold_nns = men_nns_idxs[i][men_gold_mask]
                        if within_doc:
                            gold_nns, _ = filter_by_context_doc_id(gold_nns, train_context_doc_ids[idx],
                                                                train_context_doc_ids,
                                                                return_numpy=True)
                        # If len(gold_nns) < k_men_nns then set bi_men_gold_nns[i] = [...gold_nns, -1, -1, ...]
                        bi_men_gold_nns[idx][:len(gold_nns[:k_men_nns])] = gold_nns[:k_men_nns]
            logger.info("Biencoder: Search finished")

            logger.info("Biencoder: Saving sorted biencoder train indices...")
            biencoder_train_idxs = {
                'dict_nns': bi_dict_nns,  # Nearest negative entity idxs
                'men_nns': bi_men_nns,  # Nearest negative mention idxs
                'men_gold_nns': bi_men_gold_nns  # Nearest gold mention idxs
            }
            with open(biencoder_train_idxs_pkl_path, 'wb') as write_handle:
                pickle.dump(biencoder_train_idxs, write_handle,
                            protocol=pickle.HIGHEST_PROTOCOL)
            logger.info("Biencoder: Saved")

    # Dev set
    biencoder_valid_idxs_pkl_path = os.path.join(biencoder_indices_path, 'biencoder_valid_idxs.pickle')
    if os.path.isfile(biencoder_valid_idxs_pkl_path):
        logger.info("Loading stored sorted biencoder dev indices...")
        with open(biencoder_valid_idxs_pkl_path, 'rb') as read_handle:
            biencoder_valid_idxs = pickle.load(read_handle)
    else:
        logger.info('Biencoder: Embedding and indexing valid data')
        if use_types:
            if dict_embeddings is None:
                dict_embeddings, dict_indexes, dict_idxs_by_type = data_process.embed_and_index(
                    bi_reranker, entity_dict_vecs, encoder_type="candidate", n_gpu=n_gpu, corpus=entity_dictionary,
                    force_exact_search=True, batch_size=params['embed_batch_size'])
            valid_men_embeddings, valid_men_indexes, valid_men_idxs_by_type = data_process.embed_and_index(
                bi_reranker, valid_men_vecs, encoder_type="context", n_gpu=n_gpu, corpus=valid_processed_data,
                force_exact_search=True, batch_size=params['embed_batch_size'])
        else:
            if dict_embeddings is None:
                dict_embeddings, dict_index = data_process.embed_and_index(bi_reranker, entity_dict_vecs,
                                                                           encoder_type="candidate", n_gpu=n_gpu,
                                                                           force_exact_search=True,
                                                                           batch_size=params['embed_batch_size'])
            valid_men_embeddings, valid_men_index = data_process.embed_and_index(bi_reranker, valid_men_vecs,
                encoder_type="context", n_gpu=n_gpu, force_exact_search=True, batch_size=params['embed_batch_size'])
        # valid_men_embeddings = valid_men_embeddings.numpy()
        logger.info('Biencoder: Embedding and indexing finished')

        logger.info("Biencoder: Finding nearest mentions and entities for each mention...")
        bi_dict_nns = np.zeros((len(valid_men_embeddings), k_dict_nns), dtype=int)
        bi_men_nns = -1 * np.ones((len(valid_men_embeddings), k_men_nns), dtype=int)
        if not use_types:
            _, bi_dict_nns_np = dict_index.search(valid_men_embeddings, k_dict_nns)
            n_mens_to_fetch = len(valid_men_embeddings) if within_doc else k_men_nns+1
            _, bi_men_nns_np = valid_men_index.search(valid_men_embeddings, n_mens_to_fetch)
            for i in range(len(bi_men_nns_np)):
                bi_dict_nns[i] = bi_dict_nns_np[i]
                men_nns = bi_men_nns_np[i][bi_men_nns_np[i] != i]
                if within_doc:
                    men_nns, _ = filter_by_context_doc_id(men_nns, valid_context_doc_ids[i], valid_context_doc_ids,
                                                       return_numpy=True)
                men_nns = men_nns[:k_men_nns]
                bi_men_nns[i][:len(men_nns)] = men_nns
        else:
            for entity_type in valid_men_indexes:
                men_embeds_by_type = valid_men_embeddings[valid_men_idxs_by_type[entity_type]]
                _, dict_nns_by_type = dict_indexes[entity_type].search(men_embeds_by_type, k_dict_nns)
                _, men_nns_by_type = valid_men_indexes[entity_type].search(men_embeds_by_type, len(men_embeds_by_type))
                dict_nns_idxs = np.array(list(map(lambda x: dict_idxs_by_type[entity_type][x], dict_nns_by_type)))
                men_nns_idxs = np.array(list(map(lambda x: valid_men_idxs_by_type[entity_type][x], men_nns_by_type)))
                for i, idx in enumerate(valid_men_idxs_by_type[entity_type]):
                    bi_dict_nns[idx] = dict_nns_idxs[i]
                    men_nns = men_nns_idxs[i][men_nns_idxs[i] != idx]
                    if within_doc:
                        men_nns, _ = filter_by_context_doc_id(men_nns, valid_context_doc_ids[idx],
                                                           valid_context_doc_ids,
                                                           return_numpy=True)
                    men_nns = men_nns[:k_men_nns]
                    bi_men_nns[idx][:len(men_nns)] = men_nns
        logger.info("Biencoder: Search finished")

        biencoder_valid_idxs = {
            'dict_nns': bi_dict_nns,  # Nearest entity idxs
            'men_nns': bi_men_nns  # Nearest mention idxs
        }

        logger.info("Biencoder: Saving sorted biencoder valid indices...")
        with open(biencoder_valid_idxs_pkl_path, 'wb') as write_handle:
            pickle.dump(biencoder_valid_idxs, write_handle,
                        protocol=pickle.HIGHEST_PROTOCOL)
        logger.info("Biencoder: Saved")

    return biencoder_train_idxs, biencoder_valid_idxs


def drop_entities_for_discovery_training(entity_dictionary, drop_set_fname, logger, ent_drop_prop = 0.1):
    # Drop a proportion of unique entities seen in the dev/test set from the entity dictionary
    # in order to train a new model without information leakage
    # drop_set_fname: Expects either test_processed_data.pickle or valid_process_data.pickle

    if not os.path.isfile(drop_set_fname):
        raise ValueError("Invalid --drop_set path provided to dev/test mention data")
    with open(drop_set_fname, 'rb') as read_handle:
        drop_set_data = pickle.load(read_handle)
    drop_set_mention_gold_cui_idxs = list(map(lambda x: x['label_idxs'][0], drop_set_data))
    ents_in_data = np.unique(drop_set_mention_gold_cui_idxs)
    logger.info(f"Dropping {ent_drop_prop * 100}% of {len(ents_in_data)} entities found in drop set")
    # Get entity indices to drop
    n_ents_dropped = int(ent_drop_prop * len(ents_in_data))
    rng = np.random.default_rng(seed=17)
    dropped_ent_idxs = rng.choice(ents_in_data, size=n_ents_dropped, replace=False)
    # Drop entities from dictionary
    # (subsequent call to process_mention_data will automatically drop corresponding mentions)
    keep_mask = np.ones(len(entity_dictionary), dtype='bool')
    keep_mask[dropped_ent_idxs] = False
    entity_dictionary = np.array(entity_dictionary)[keep_mask]
    return entity_dictionary


def load_training_data(bi_tokenizer,
                       max_context_length,
                       max_cand_length,
                       knn,
                       pickle_src_path,
                       params,
                       logger,
                       return_dict_only=False):
    # Load training data
    entity_dictionary_loaded = False
    entity_dictionary_pkl_path = os.path.join(pickle_src_path, 'entity_dictionary.pickle')
    if os.path.isfile(entity_dictionary_pkl_path):
        print("Loading stored processed entity dictionary...")
        with open(entity_dictionary_pkl_path, 'rb') as read_handle:
            entity_dictionary = pickle.load(read_handle)
        entity_dictionary_loaded = True

    if return_dict_only and entity_dictionary_loaded:
        return entity_dictionary

    # Load train data
    train_tensor_data_pkl_path = os.path.join(pickle_src_path, 'train_tensor_data.pickle')
    train_processed_data_pkl_path = os.path.join(pickle_src_path, 'train_processed_data.pickle')
    if not params["drop_entities"] and os.path.isfile(train_tensor_data_pkl_path) and os.path.isfile(train_processed_data_pkl_path):
        print("Loading stored processed train data...")
        with open(train_tensor_data_pkl_path, 'rb') as read_handle:
            train_tensor_data = pickle.load(read_handle)
        with open(train_processed_data_pkl_path, 'rb') as read_handle:
            train_processed_data = pickle.load(read_handle)
    else:
        train_samples = read_dataset("train", params["data_path"])
        if not entity_dictionary_loaded:
            with open(os.path.join(params["data_path"], 'dictionary.pickle'), 'rb') as read_handle:
                entity_dictionary = pickle.load(read_handle)

        # Check if dataset has multiple ground-truth labels
        mult_labels = "labels" in train_samples[0].keys()
        # Filter samples without gold entities
        train_samples = list(
            filter(lambda sample: (len(sample["labels"]) > 0) if mult_labels else (sample["label"] is not None),
                   train_samples))
        logger.info("Read %d train samples." % len(train_samples))

        # For discovery experiment: Drop entities used in training that were dropped randomly from dev/test set
        if params["drop_entities"]:
            assert entity_dictionary is not None
            entity_dictionary = drop_entities_for_discovery_training(entity_dictionary,
                                                                     params["drop_set"],
                                                                     logger)

        train_processed_data, entity_dictionary, train_tensor_data = data_process.process_mention_data(
            train_samples,
            entity_dictionary,
            bi_tokenizer,
            max_context_length,
            max_cand_length,
            context_key=params["context_key"],
            multi_label_key="labels" if mult_labels else None,
            silent=params["silent"],
            logger=logger,
            debug=params["debug"],
            knn=knn,
            dictionary_processed=entity_dictionary_loaded
        )
        print("Saving processed train data...")
        if not entity_dictionary_loaded:
            with open(entity_dictionary_pkl_path, 'wb') as write_handle:
                pickle.dump(entity_dictionary, write_handle,
                            protocol=pickle.HIGHEST_PROTOCOL)
        with open(train_tensor_data_pkl_path, 'wb') as write_handle:
            pickle.dump(train_tensor_data, write_handle,
                        protocol=pickle.HIGHEST_PROTOCOL)
        with open(train_processed_data_pkl_path, 'wb') as write_handle:
            pickle.dump(train_processed_data, write_handle,
                        protocol=pickle.HIGHEST_PROTOCOL)

    if return_dict_only:
        return entity_dictionary
    return entity_dictionary, train_tensor_data, train_processed_data


def load_validation_data(bi_tokenizer,
                         max_context_length,
                         max_cand_length,
                         entity_dictionary,
                         knn,
                         pickle_src_path,
                         params,
                         logger):
    valid_tensor_data_pkl_path = os.path.join(pickle_src_path, 'valid_tensor_data.pickle')
    valid_processed_data_pkl_path = os.path.join(pickle_src_path, 'valid_processed_data.pickle')
    if os.path.isfile(valid_tensor_data_pkl_path) and os.path.isfile(valid_processed_data_pkl_path):
        print("Loading stored processed valid data...")
        with open(valid_tensor_data_pkl_path, 'rb') as read_handle:
            valid_tensor_data = pickle.load(read_handle)
        with open(valid_processed_data_pkl_path, 'rb') as read_handle:
            valid_processed_data = pickle.load(read_handle)
    else:
        valid_samples = read_dataset("valid", params["data_path"])
        # Check if dataset has multiple ground-truth labels
        mult_labels = "labels" in valid_samples[0].keys()
        # Filter samples without gold entities
        valid_samples = list(
            filter(lambda sample: (len(sample["labels"]) > 0) if mult_labels else (sample["label"] is not None),
                   valid_samples))
        logger.info("Read %d valid samples." % len(valid_samples))

        valid_processed_data, _, valid_tensor_data = data_process.process_mention_data(
            valid_samples,
            entity_dictionary,
            bi_tokenizer,
            max_context_length,
            max_cand_length,
            context_key=params["context_key"],
            multi_label_key="labels" if mult_labels else None,
            silent=params["silent"],
            logger=logger,
            debug=params["debug"],
            knn=knn,
            dictionary_processed=True
        )
        print("Saving processed valid data...")
        with open(valid_tensor_data_pkl_path, 'wb') as write_handle:
            pickle.dump(valid_tensor_data, write_handle,
                        protocol=pickle.HIGHEST_PROTOCOL)
        with open(valid_processed_data_pkl_path, 'wb') as write_handle:
            pickle.dump(valid_processed_data, write_handle,
                        protocol=pickle.HIGHEST_PROTOCOL)
    return valid_tensor_data, valid_processed_data


def get_gold_clusters(train_processed_data):
    train_gold_clusters = data_process.compute_gold_clusters(train_processed_data)
    max_gold_cluster_len = 0
    for ent in train_gold_clusters:
        if len(train_gold_clusters[ent]) > max_gold_cluster_len:
            max_gold_cluster_len = len(train_gold_clusters[ent])
    return train_gold_clusters, max_gold_cluster_len


def get_gold_arbo_links(cross_reranker,
                        max_context_length,
                        entity_dict_vecs,
                        train_men_vecs,
                        train_processed_data,
                        gold_men_nns,
                        max_seq_length,
                        knn,
                        within_doc=False,
                        train_context_doc_ids=None,
                        debug=False):
    cross_reranker.model.eval()
    with torch.no_grad():
        gold_links = {}
        n_entities, n_mentions = len(entity_dict_vecs), len(train_men_vecs)
        _iter = tqdm(range(len(train_men_vecs) if not debug else 200), desc="Computing gold arbos")
        for mention_idx in _iter:
            # Assuming that there is only 1 gold label
            cluster_ent = train_processed_data[mention_idx]['label_idxs'][0]
            if mention_idx not in gold_links:
                # Get <knn> nearest (biencoder) gold mentions
                gold_nns = gold_men_nns[mention_idx][gold_men_nns[mention_idx] != -1]
                if within_doc:
                    gold_nns, _ = filter_by_context_doc_id(gold_nns,
                                                        train_context_doc_ids[mention_idx],
                                                        train_context_doc_ids,
                                                        return_numpy=True)
                cluster_mens = np.concatenate(([mention_idx], gold_nns[:knn]))
                # Simply link to the entity if cluster is singleton
                if len(cluster_mens) == 1:
                    gold_links[mention_idx] = cluster_ent
                    continue
                # Run MST on the current query mention's (biencoder) k-NN gold sub-cluster to find positive edges
                rows, cols, data, shape = [], [], [], (n_entities + n_mentions, n_entities + n_mentions)
                to_ent_input = concat_for_crossencoder(train_men_vecs[cluster_mens],
                                                       entity_dict_vecs[cluster_ent].expand(len(cluster_mens), 1,
                                                                                            entity_dict_vecs.size(1)),
                                                       max_seq_length)  # Shape: N x 1 x 2D
                try:
                    to_ent_data = score_in_batches(cross_reranker, max_context_length, to_ent_input,
                                                   is_context_encoder=False, silent=True)
                except:
                    raise ValueError(f"Probably a batch size error. The current cluster size was {len(cluster_mens)}")
                to_men_input = concat_for_crossencoder(train_men_vecs[cluster_mens],
                                                       train_men_vecs[cluster_mens].expand(len(cluster_mens),
                                                                                           len(cluster_mens),
                                                                                           train_men_vecs.size(1)),
                                                       max_seq_length)  # Shape: N x N x 2D
                try:
                    to_men_data = score_in_batches(cross_reranker, max_context_length, to_men_input,
                                                   is_context_encoder=True, silent=True)
                except:
                    raise ValueError(f"Probably a batch size error. The current cluster size was {len(cluster_mens)}")
                for i in range(len(cluster_mens)):
                    from_node = n_entities + cluster_mens[i]
                    to_node = cluster_ent
                    # Add mention-entity link
                    rows.append(from_node)
                    cols.append(to_node)
                    data.append(to_ent_data[i, 0].item())
                    # Add mention-mention links
                    for j in range(len(cluster_mens)):
                        # Skip diagonal elements from to_men_data because they are scores of elements with themselves
                        if i == j:
                            continue
                        to_node = n_entities + cluster_mens[j]
                        rows.append(from_node)
                        cols.append(to_node)
                        data.append(to_men_data[i, j].item())
                # Partition graph with entity constraint
                rows, cols, data = cluster_linking_partition(np.array(rows),
                                                             np.array(cols),
                                                             np.array(data),
                                                             n_entities,
                                                             directed=True,
                                                             silent=True)
                assert np.array_equal(np.sort(rows - n_entities), np.sort(cluster_mens))
                for i in range(len(rows)):
                    men_idx = rows[i] - n_entities
                    assert men_idx >= 0
                    if men_idx in gold_links:
                        continue
                    # Only add gold link for current mention's arborescence computation
                    if men_idx == mention_idx:
                        gold_links[men_idx] = cols[i]
                        break
    return gold_links


def construct_train_batch(cross_reranker,
                          mention_idxs,
                          gold_links,
                          entity_dict_vecs,
                          train_men_vecs,
                          neg_ent_topk_inputs,
                          neg_men_topk_inputs,
                          max_seq_length,
                          max_context_length):
    n_entities = len(entity_dict_vecs)
    batch_positive_scores = None
    # Construct the batch matrix
    for mention_idx in mention_idxs:
        mention_idx = mention_idx.item()
        # Forward-propagate the positive edge
        # Case: Gold edge is an entity
        if gold_links[mention_idx] < n_entities:
            is_context_encoder = False
            to_idx = gold_links[mention_idx]
            to_input = torch.unsqueeze(entity_dict_vecs[to_idx:to_idx + 1], dim=0)
        # Case: Gold edge is a mention
        else:
            is_context_encoder = True
            to_idx = gold_links[mention_idx] - n_entities
            to_input = torch.unsqueeze(train_men_vecs[to_idx:to_idx + 1], dim=0)
        pos_label_input = concat_for_crossencoder(train_men_vecs[mention_idx:mention_idx + 1],
                                                  to_input,
                                                  max_seq_length)  # Shape: 1 x 1 x 2D
        pos_score = score_in_batches(cross_reranker, max_context_length, pos_label_input.cuda(),
                                     is_context_encoder=is_context_encoder)
        batch_positive_scores = pos_score if batch_positive_scores is None else \
            torch.cat((batch_positive_scores, pos_score), dim=0)
    batch_negative_ent_inputs = neg_ent_topk_inputs[mention_idxs]
    batch_negative_men_inputs = neg_men_topk_inputs[mention_idxs]

    return batch_positive_scores, batch_negative_ent_inputs.cuda(), batch_negative_men_inputs.cuda()


def build_cross_concat_input(biencoder_idxs,
                             men_vecs,
                             entity_dict_vecs,
                             max_seq_length,
                             k_biencoder):
    men_men_inputs = []
    men_ent_inputs = []
    for mention_idx in tqdm(range(len(men_vecs))):
        # Get nearest biencoder mentions and entities
        bi_men_idxs = biencoder_idxs['men_nns'][mention_idx][:k_biencoder]
        bi_ent_idxs = biencoder_idxs['dict_nns'][mention_idx][:k_biencoder]
        bi_men_inputs = None
        for bi_men_idx in bi_men_idxs:
            bi_men_inputs = men_vecs[bi_men_idx].unsqueeze(dim=0) if bi_men_inputs is None else \
                torch.cat((bi_men_inputs, men_vecs[bi_men_idx].unsqueeze(dim=0)), dim=0)
        bi_men_inputs = torch.unsqueeze(bi_men_inputs, dim=0)
        bi_ent_inputs = None
        for bi_ent_idx in bi_ent_idxs:
            bi_ent_inputs = entity_dict_vecs[bi_ent_idx].unsqueeze(dim=0) if bi_ent_inputs is None else \
                torch.cat((bi_ent_inputs, entity_dict_vecs[bi_ent_idx].unsqueeze(dim=0)), dim=0)
        bi_ent_inputs = torch.unsqueeze(bi_ent_inputs, dim=0)
        # Concatenate for cross-encoder
        cross_men_inputs = concat_for_crossencoder(men_vecs[mention_idx:mention_idx + 1],
                                                   bi_men_inputs,
                                                   max_seq_length)  # Shape: 1 x knn x 2D
        cross_ent_inputs = concat_for_crossencoder(men_vecs[mention_idx:mention_idx + 1],
                                                   bi_ent_inputs,
                                                   max_seq_length)  # Shape: 1 x knn x 2D
        men_men_inputs.append(cross_men_inputs)
        men_ent_inputs.append(cross_ent_inputs)
    men_men_inputs = torch.cat(men_men_inputs)
    men_ent_inputs = torch.cat(men_ent_inputs)
    return men_men_inputs, men_ent_inputs


def get_train_neg_cross_inputs(cross_reranker,
                               max_context_length,
                               train_men_concat_inputs,
                               train_ent_concat_inputs,
                               n_knn_men_negs,
                               bi_train_nn_count,
                               n_knn_ent_negs,
                               logger,
                               debug=False):
    def _helper(concat_inputs, n_knn, is_context_encoder):
        if debug:
            concat_inputs = concat_inputs[:200]
        scores = score_in_batches(cross_reranker, max_context_length, concat_inputs,
                                  is_context_encoder=is_context_encoder)
        if is_context_encoder:
            for i in range(len(scores)):
                # Set scores for all invalid nearest neighbours to -infinity (due to variable NN counts of mentions)
                scores[i][bi_train_nn_count[i]:] = float('-inf')
        topk_idxs = torch.argsort(scores, dim=1, descending=True)[:, :n_knn]
        stacked = []
        for r in range(topk_idxs.size(0)):
            stacked.append(concat_inputs[r][topk_idxs[r]])
        topk_inputs = torch.stack(stacked)
        logger.info('Scoring done')
        return topk_inputs

    # Score biencoder negatives using cross-encoder and store nearest-k for current epoch
    cross_reranker.model.eval()
    with torch.no_grad():
        logger.info('Scoring mention-mention negative edges using cross-encoder...')
        neg_men_topk_inputs = _helper(concat_inputs=train_men_concat_inputs, n_knn=n_knn_men_negs,
                                      is_context_encoder=True)
        logger.info('Scoring mention-entity negative edges using cross-encoder...')
        neg_ent_topk_inputs = _helper(concat_inputs=train_ent_concat_inputs, n_knn=n_knn_ent_negs,
                                      is_context_encoder=False)
    return neg_men_topk_inputs, neg_ent_topk_inputs


def execute_training_step(mention_idxs,
                          bi_train_nn_count,
                          n_knn_men_negs,
                          batch_positive_scores,
                          batch_negative_men_inputs,
                          batch_negative_ent_inputs,
                          cross_reranker,
                          max_context_length,
                          grad_acc_steps,
                          n_gpu,
                          no_sigmoid=False):
    # Handle variable mention-negatives lengths for different training queries in the batch
    min_negs = float('inf')
    dual_negs_mask = np.array([True] * len(mention_idxs))  # Mask to keep queries with both ent and men negs
    skipped = 0
    for i, m_idx in enumerate(mention_idxs):
        if bi_train_nn_count[m_idx] == 0:  # If no negative mentions neighbours
            dual_negs_mask[i] = False
            skipped += 1
            continue
        min_negs = min(min_negs, bi_train_nn_count[m_idx], n_knn_men_negs)
    loss_dual_negs = loss_ent_negs = 0
    assert min_negs != float('inf') or skipped > 0
    if min_negs != float('inf'):  # i.e. at least 1 query has dual negs
        positive_scores = batch_positive_scores[dual_negs_mask]
        negative_men_inputs = batch_negative_men_inputs[dual_negs_mask][:, :min_negs]
        negative_ent_inputs = batch_negative_ent_inputs[dual_negs_mask]
        # FIX: for error scenario of less number of examples than number of GPUs while using Data Parallel
        data_parallel_batch_size_check = negative_men_inputs.shape[0] * negative_men_inputs.shape[1] >= n_gpu and \
                                         negative_ent_inputs.shape[0] * negative_ent_inputs.shape[1] >= n_gpu
        if data_parallel_batch_size_check:
            loss_dual_negs = cross_reranker(positive_scores, negative_men_inputs,
                                            negative_ent_inputs, max_context_length,
                                            no_sigmoid=no_sigmoid)
    if skipped > 0:
        positive_scores = batch_positive_scores[~dual_negs_mask]
        negative_men_inputs = None
        negative_ent_inputs = batch_negative_ent_inputs[~dual_negs_mask]
        data_parallel_batch_size_check = negative_ent_inputs.shape[0] * negative_ent_inputs.shape[1] >= n_gpu
        if data_parallel_batch_size_check:
            loss_ent_negs = cross_reranker(positive_scores, negative_men_inputs,
                                           negative_ent_inputs, max_context_length,
                                           no_sigmoid=no_sigmoid)
    loss = ((loss_dual_negs * (len(mention_idxs) - skipped) + loss_ent_negs * skipped) / len(mention_idxs)) / grad_acc_steps
    loss.backward()
    return loss.item()


def main(params):
    # Parameter initializations
    logger = setup_logger()
    debug = params["debug"]
    global SCORING_BATCH_SIZE
    SCORING_BATCH_SIZE = params["scoring_batch_size"]
    model_output_path = params["output_path"]
    if not os.path.exists(model_output_path):
        os.makedirs(model_output_path)
    pickle_src_path = params["pickle_src_path"]
    if pickle_src_path is None or not os.path.exists(pickle_src_path):
        pickle_src_path = model_output_path
    biencoder_indices_path = params["biencoder_indices_path"]
    if biencoder_indices_path is None:
        biencoder_indices_path = model_output_path
    elif not os.path.exists(biencoder_indices_path):
        os.makedirs(biencoder_indices_path)
    knn = params["knn"]  # Number of (positive+negatives) in each row of the training batch
    use_types = params["use_types"]
    within_doc = params["within_doc"]

    # Bi-encoder model
    biencoder_params = copy.deepcopy(params)
    biencoder_params['add_linear'] = False
    bi_reranker = BiEncoderRanker(biencoder_params)
    bi_tokenizer = bi_reranker.tokenizer
    bi_knn = params["bi_knn"]  # Number of biencoder nearest-neighbors to fetch for cross-encoder scoring (default: 64)

    # Cross-encoder model
    cross_reranker = CrossEncoderRanker(params)
    cross_model = cross_reranker.model
    cross_tokenizer = cross_reranker.tokenizer
    device = cross_reranker.device
    n_gpu = cross_reranker.n_gpu
    no_sigmoid_train = params["no_sigmoid_train"]

    # Training params
    grad_acc_steps = params["gradient_accumulation_steps"]
    if grad_acc_steps < 1:
        raise ValueError("Invalid gradient_accumulation_steps parameter: {}, should be >= 1".format(grad_acc_steps))
    # Gradient accumulation:
    # Model's batch size `z` = `x / y`. Effective batch size of `x` when accumulating gradients across `y` batches
    params["train_batch_size"] = (
        params["train_batch_size"] // grad_acc_steps
    )
    train_batch_size = params["train_batch_size"]

    # Input lengths
    max_seq_length = params["max_seq_length"]
    max_context_length = params["max_context_length"]
    max_cand_length = params["max_cand_length"]

    # Fix random seeds
    seed = params["seed"]
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    if cross_reranker.n_gpu > 0:
        torch.cuda.manual_seed_all(seed)

    # Load data
    if params["only_evaluate"]:
        entity_dictionary = load_training_data(bi_tokenizer,
                                               max_context_length,
                                               max_cand_length,
                                               knn,
                                               pickle_src_path,
                                               params,
                                               logger,
                                               return_dict_only=True)
    else:
        # Load training data
        entity_dictionary, train_tensor_data, train_processed_data = load_training_data(bi_tokenizer,
                                                                                        max_context_length,
                                                                                        max_cand_length,
                                                                                        knn,
                                                                                        pickle_src_path,
                                                                                        params,
                                                                                        logger)
        # Store query mention vectors
        train_men_vecs = train_tensor_data[:][0]
    # Store entity dictionary vectors
    entity_dict_vecs = torch.tensor(list(map(lambda x: x['ids'], entity_dictionary)), dtype=torch.long)

    # Load validation data
    valid_tensor_data, valid_processed_data = load_validation_data(bi_tokenizer,
                                                                   max_context_length,
                                                                   max_cand_length,
                                                                   entity_dictionary,
                                                                   knn,
                                                                   pickle_src_path,
                                                                   params,
                                                                   logger)
    # Store query mention vectors
    valid_men_vecs = valid_tensor_data[:][0]

    train_context_doc_ids = valid_context_doc_ids = None
    if within_doc:
        # Get context_document_ids for each mention in training and validation
        train_context_doc_ids = get_context_doc_ids("train", params)
        valid_context_doc_ids = get_context_doc_ids("valid", params)

    if params["only_evaluate"]:
        _, biencoder_valid_idxs = get_biencoder_nns(bi_reranker=bi_reranker,
                                                    biencoder_indices_path=biencoder_indices_path,
                                                    entity_dictionary=entity_dictionary,
                                                    entity_dict_vecs=entity_dict_vecs,
                                                    train_men_vecs=None,
                                                    train_processed_data=None,
                                                    train_gold_clusters=None,
                                                    valid_men_vecs=valid_men_vecs,
                                                    valid_processed_data=valid_processed_data,
                                                    use_types=use_types,
                                                    logger=logger,
                                                    n_gpu=n_gpu,
                                                    params=params,
                                                    train_context_doc_ids=train_context_doc_ids,
                                                    valid_context_doc_ids=valid_context_doc_ids)
        bi_valid_nn_count = np.sum(biencoder_valid_idxs['men_nns'] != -1, axis=1)
        # Compute and store the concatenated cross-encoder inputs for validation
        valid_men_concat_inputs, valid_ent_concat_inputs = build_cross_concat_input(biencoder_valid_idxs,
                                                                                    valid_men_vecs,
                                                                                    entity_dict_vecs,
                                                                                    max_seq_length,
                                                                                    bi_knn)
        evaluate(cross_reranker,
                 max_context_length,
                 entity_dictionary,
                 valid_processed_data,
                 biencoder_valid_idxs,
                 bi_valid_nn_count,
                 valid_men_concat_inputs,
                 valid_ent_concat_inputs,
                 logger,
                 max_k=8,
                 k_biencoder=64)
        exit()

    # Get clusters of gold mentions that map to a unique entity
    train_gold_clusters, max_gold_cluster_len = get_gold_clusters(train_processed_data)

    # Get indices of nearest mentions and entities from the trained biencoder
    # (to reduce candidates to score by cross-encoder)
    biencoder_train_idxs, biencoder_valid_idxs = get_biencoder_nns(bi_reranker=bi_reranker,
                                                                   biencoder_indices_path=biencoder_indices_path,
                                                                   entity_dictionary=entity_dictionary,
                                                                   entity_dict_vecs=entity_dict_vecs,
                                                                   train_men_vecs=train_men_vecs,
                                                                   train_processed_data=train_processed_data,
                                                                   train_gold_clusters=train_gold_clusters,
                                                                   valid_men_vecs=valid_men_vecs,
                                                                   valid_processed_data=valid_processed_data,
                                                                   use_types=use_types,
                                                                   logger=logger,
                                                                   n_gpu=n_gpu,
                                                                   params=params,
                                                                   train_context_doc_ids=train_context_doc_ids,
                                                                   valid_context_doc_ids=valid_context_doc_ids
                                                                   )
    bi_train_nn_count = np.sum(biencoder_train_idxs['men_nns'] != -1, axis=1)
    bi_valid_nn_count = np.sum(biencoder_valid_idxs['men_nns'] != -1, axis=1)

    # Compute and store the concatenated cross-encoder inputs for validation
    logger.info('Computing the concatenated cross-encoder inputs for validation...')
    valid_men_concat_inputs, valid_ent_concat_inputs = build_cross_concat_input(biencoder_valid_idxs,
                                                                                valid_men_vecs,
                                                                                entity_dict_vecs,
                                                                                max_seq_length,
                                                                                bi_knn)
    logger.info('Done')
    # Compute and store the concatenated cross-encoder inputs for the negatives in training
    logger.info('Computing the concatenated cross-encoder negative inputs for training...')
    train_men_concat_inputs, train_ent_concat_inputs = build_cross_concat_input(biencoder_train_idxs,
                                                                                train_men_vecs,
                                                                                entity_dict_vecs,
                                                                                max_seq_length,
                                                                                bi_knn)
    logger.info('Done')

    if not params["skip_initial_eval"]:
        logger.info("Evaluating dev set on untrained model...")
        # Evaluate cross-encoder before training
        evaluate(cross_reranker,
                 max_context_length,
                 entity_dictionary,
                 valid_processed_data,
                 biencoder_valid_idxs,
                 bi_valid_nn_count,
                 valid_men_concat_inputs,
                 valid_ent_concat_inputs,
                 logger,
                 max_k=8,
                 k_biencoder=64)

    # Reduce dataset size for quick debugging
    if debug:
        train_tensor_data = train_tensor_data[:200]

    # Initialize training data loader
    train_sampler = RandomSampler(train_tensor_data) if params["shuffle"] else SequentialSampler(train_tensor_data)
    train_dataloader = DataLoader(train_tensor_data, sampler=train_sampler, batch_size=train_batch_size)
    if not params["silent"]:
        train_dataloader = tqdm(train_dataloader, desc="Batch")

    # Start training
    time_start = time.time()
    write_to_file(
        os.path.join(model_output_path, "training_params.txt"), str(params)
    )
    checkpoint_pkl_path = os.path.join(model_output_path, 'epoch_checkpoint.pickle')
    logger.info("Starting training")
    logger.info(
        "device: {} n_gpu: {}, data_parallel: {}".format(device, n_gpu, params["data_parallel"])
    )

    optim_sched, optimizer, scheduler = load_optimizer_scheduler(params, logger), None, None
    if optim_sched is None:
        optimizer = get_optimizer(cross_model, params)
        scheduler = get_scheduler(params, optimizer, len(train_tensor_data), logger)
    else:
        optimizer = optim_sched['optimizer']
        scheduler = optim_sched['scheduler']

    best_score = {
        'directed': {
            'acc': -1.,
            'epoch': 0,
            'k': 0
        },
        'undirected': {
            'acc': -1.,
            'epoch': 0,
            'k': 0
        }
    }
    n_knn_negs = params["knn_negs"]  # Number of k-NN negatives in each row of the training batch (default: 8)
    n_knn_ent_negs, n_knn_men_negs = n_knn_negs // 2, n_knn_negs // 2

    for epoch_idx in range(params["num_train_epochs"]):
        logger.info(f"***** Starting EPOCH {epoch_idx}/{params['num_train_epochs']-1} *****")
        torch.cuda.empty_cache()

        checkpoint_data = None
        if params["checkpoint_epoch_data"]:
            if os.path.isfile(checkpoint_pkl_path):
                logger.info("Loading stored epoch checkpoint data...")
                with open(checkpoint_pkl_path, 'rb') as read_handle:
                    checkpoint_data = pickle.load(read_handle)
                    gold_links = checkpoint_data['gold_links']
                    neg_men_topk_inputs = checkpoint_data['neg_men_topk_inputs']
                    neg_ent_topk_inputs = checkpoint_data['neg_ent_topk_inputs']

        if checkpoint_data is None:
            # Compute arborescences per gold k-NN cluster for each mention and store the ground-truth positive edges
            logger.info("Computing gold arborescence links for positive training labels")
            gold_links = get_gold_arbo_links(cross_reranker,
                                             max_context_length,
                                             entity_dict_vecs,
                                             train_men_vecs,
                                             train_processed_data,
                                             biencoder_train_idxs['men_gold_nns'],
                                             max_seq_length,
                                             knn=params.get("gold_arbo_knn", 1),
                                             debug=debug,
                                             within_doc=within_doc,
                                             train_context_doc_ids=train_context_doc_ids)
            logger.info("Done")

            # Score biencoder negatives using cross-encoder and store nearest-k for this epoch
            logger.info("Computing cross-encoder inputs for hard-negatives")
            neg_men_topk_inputs, neg_ent_topk_inputs = get_train_neg_cross_inputs(cross_reranker,
                                                                                  max_context_length,
                                                                                  train_men_concat_inputs,
                                                                                  train_ent_concat_inputs,
                                                                                  n_knn_men_negs,
                                                                                  bi_train_nn_count,
                                                                                  n_knn_ent_negs,
                                                                                  logger,
                                                                                  debug=debug)
            logger.info("Done")

        if params["checkpoint_epoch_data"] and checkpoint_data is None:
            logger.info("Checkpointing epoch data...")
            with open(checkpoint_pkl_path, 'wb') as write_handle:
                pickle.dump({
                    'gold_links': gold_links,
                    'neg_men_topk_inputs': neg_men_topk_inputs,
                    'neg_ent_topk_inputs': neg_ent_topk_inputs
                }, write_handle, protocol=pickle.HIGHEST_PROTOCOL)
            logger.info("Checkpoint saved!")

        save_intervals_done = 0
        tr_loss = 0
        cross_model.train()
        for step, batch in enumerate(train_dataloader):
            _, _, _, mention_idxs = batch

            batch_positive_scores, batch_negative_ent_inputs, batch_negative_men_inputs = construct_train_batch(
                cross_reranker=cross_reranker,
                mention_idxs=mention_idxs,
                gold_links=gold_links,
                entity_dict_vecs=entity_dict_vecs,
                train_men_vecs=train_men_vecs,
                neg_ent_topk_inputs=neg_ent_topk_inputs,
                neg_men_topk_inputs=neg_men_topk_inputs,
                max_seq_length=max_seq_length,
                max_context_length=max_context_length)

            total_loss = execute_training_step(mention_idxs,
                                               bi_train_nn_count,
                                               n_knn_men_negs,
                                               batch_positive_scores,
                                               batch_negative_men_inputs,
                                               batch_negative_ent_inputs,
                                               cross_reranker,
                                               max_context_length,
                                               grad_acc_steps,
                                               n_gpu,
                                               no_sigmoid=no_sigmoid_train)
            tr_loss += total_loss

            n_print_iters = params["print_interval"] * grad_acc_steps
            if (step + 1) % n_print_iters == 0:
                logger.info(
                    "Step {} - epoch {} average loss: {}".format(
                        step,
                        epoch_idx,
                        tr_loss / n_print_iters,
                    )
                )
                tr_loss = 0

            if (step + 1) % grad_acc_steps == 0:
                torch.nn.utils.clip_grad_norm_(
                    cross_model.parameters(), params["max_grad_norm"]
                )
                optimizer.step()
                scheduler.step()
                optimizer.zero_grad()

            if params["eval_interval"] != -1:
                if (step + 1) % (params["eval_interval"] * grad_acc_steps) == 0:
                    logger.info("Evaluation on the dev set")
                    evaluate(cross_reranker,
                             max_context_length,
                             entity_dictionary,
                             valid_processed_data,
                             biencoder_valid_idxs,
                             bi_valid_nn_count,
                             valid_men_concat_inputs,
                             valid_ent_concat_inputs,
                             logger,
                             max_k=8,
                             k_biencoder=64)
                    cross_model.train()
                    logger.info("\n")

            if params["save_interval"] != -1:
                train_prop_done = (step + 1) / len(train_dataloader)
                # Skip saving the model at 100% completion since we do that regardless in the outer loop
                if train_prop_done != 1 and train_prop_done >= (save_intervals_done + 1) * params["save_interval"]:
                    epoch_output_folder_path = os.path.join(
                        model_output_path, f"epoch_{epoch_idx}_{save_intervals_done}"
                    )
                    save_intervals_done += 1
                    logger.info(
                        f"***** Saving fine-tuned model (at {save_intervals_done * params['save_interval'] * 100}%) *****")
                    save_model(cross_model, cross_tokenizer, epoch_output_folder_path)
                    logger.info(f"Model saved at {epoch_output_folder_path}")

        logger.info("***** Saving fine-tuned model *****")
        epoch_output_folder_path = os.path.join(
            model_output_path, "epoch_{}".format(epoch_idx)
        )
        save_model(cross_model, cross_tokenizer, epoch_output_folder_path, scheduler, optimizer)
        logger.info(f"Model saved at {epoch_output_folder_path}")

        if not params["skip_epoch_eval"]:
            eval_accuracy = evaluate(cross_reranker,
                                     max_context_length,
                                     entity_dictionary,
                                     valid_processed_data,
                                     biencoder_valid_idxs,
                                     bi_valid_nn_count,
                                     valid_men_concat_inputs,
                                     valid_ent_concat_inputs,
                                     logger,
                                     max_k=8,
                                     k_biencoder=64)
            for mode in ['directed', 'undirected']:
                ls = [best_score[mode]['acc'], eval_accuracy[mode]]
                li = [best_score[mode]['epoch'], epoch_idx]
                best_score[mode]['acc'] = ls[np.argmax(ls)]
                best_score[mode]['epoch'] = li[np.argmax(ls)]
                if np.argmax(ls) == 0:
                    best_score[mode]['k'] = eval_accuracy[f'{mode}_k']
            logger.info("\n")

        if params["checkpoint_epoch_data"] and epoch_idx < params["num_train_epochs"] - 1:
            if os.path.isfile(checkpoint_pkl_path):
                logger.info("Clearing stored epoch checkpoint data...")
                os.remove(checkpoint_pkl_path)

    execution_time = (time.time() - time_start) / 60
    logger.info("The training took {} minutes".format(execution_time))
    logger.info("Best performance:")
    logger.info(json.dumps(best_score))
    logger.info("***** Training finished *****")


if __name__ == "__main__":
    parser = BlinkParser(add_model_args=True)
    parser.add_training_args()
    parser.add_joint_train_args()
    args = parser.parse_args()
    print(args)
    main(args.__dict__)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import math
from typing import Dict, List, Optional

import torch
import torch.nn as nn
import torch.nn.functional as F
from fairseq import search, utils
from fairseq.data import data_utils
from fairseq.models import FairseqIncrementalDecoder
from torch import Tensor
from fairseq.ngram_repeat_block import NGramRepeatBlock


class SequenceGenerator(nn.Module):
    def __init__(
        self,
        models,
        tgt_dict,
        beam_size=1,
        max_len_a=0,
        max_len_b=200,
        min_len=1,
        normalize_scores=True,
        len_penalty=1.0,
        unk_penalty=0.0,
        temperature=1.0,
        match_source_len=False,
        no_repeat_ngram_size=0,
        search_strategy=None,
        eos=None,
        symbols_to_strip_from_output=None,
        lm_model=None,
        lm_weight=1.0,
    ):
        """Generates translations of a given source sentence.

        Args:
            models (List[~fairseq.models.FairseqModel]): ensemble of models,
                currently support fairseq.models.TransformerModel for scripting
            beam_size (int, optional): beam width (default: 1)
            max_len_a/b (int, optional): generate sequences of maximum length
                ax + b, where x is the source length
            min_len (int, optional): the minimum length of the generated output
                (not including end-of-sentence)
            normalize_scores (bool, optional): normalize scores by the length
                of the output (default: True)
            len_penalty (float, optional): length penalty, where <1.0 favors
                shorter, >1.0 favors longer sentences (default: 1.0)
            unk_penalty (float, optional): unknown word penalty, where <0
                produces more unks, >0 produces fewer (default: 0.0)
            temperature (float, optional): temperature, where values
                >1.0 produce more uniform samples and values <1.0 produce
                sharper samples (default: 1.0)
            match_source_len (bool, optional): outputs should match the source
                length (default: False)
        """
        super().__init__()
        if isinstance(models, EnsembleModel):
            self.model = models
        else:
            self.model = EnsembleModel(models)
        self.tgt_dict = tgt_dict
        self.pad = self.model.single_model.config.pad_token_id
        self.unk = None # hf Bart has no unk token id
        self.eos = self.model.single_model.config.eos_token_id
        self.symbols_to_strip_from_output = (
            symbols_to_strip_from_output.union({self.eos})
            if symbols_to_strip_from_output is not None
            else {self.eos}
        )
        self.vocab_size = self.model.single_model.config.vocab_size
        self.beam_size = beam_size
        # the max beam size is the dictionary size - 1, since we never select pad
        self.beam_size = min(beam_size, self.vocab_size - 1)
        self.max_len_a = max_len_a
        self.max_len_b = max_len_b
        self.min_len = min_len

        self.normalize_scores = normalize_scores
        self.len_penalty = len_penalty
        self.unk_penalty = unk_penalty
        self.temperature = temperature
        self.match_source_len = match_source_len

        if no_repeat_ngram_size > 0:
            self.repeat_ngram_blocker = NGramRepeatBlock(no_repeat_ngram_size)
        else:
            self.repeat_ngram_blocker = None

        assert temperature > 0, "--temperature must be greater than 0"

        self.search = (
            search.BeamSearch(tgt_dict) if search_strategy is None else search_strategy
        )
        # We only need to set src_lengths in LengthConstrainedBeamSearch.
        # As a module attribute, setting it would break in multithread
        # settings when the model is shared.
        self.should_set_src_lengths = (
            hasattr(self.search, "needs_src_lengths") and self.search.needs_src_lengths
        )

        self.model.single_model.eval()

        self.lm_model = lm_model
        self.lm_weight = lm_weight
        if self.lm_model is not None:
            self.lm_model.eval()

    def cuda(self):
        self.model.cuda()
        return self

    @torch.no_grad()
    def forward(
        self,
        sample: Dict[str, Dict[str, Tensor]],
        prefix_tokens: Optional[Tensor] = None,
        bos_token: Optional[int] = None,
        prefix_mention_is: bool = False,
    ):
        """Generate a batch of translations.

        Args:
            sample (dict): batch
            prefix_tokens (torch.LongTensor, optional): force decoder to begin
                with these tokens
            bos_token (int, optional): beginning of sentence token
                (default: self.eos)
        """
        return self._generate(sample, prefix_tokens, bos_token=bos_token, prefix_mention_is=prefix_mention_is)

    # TODO(myleott): unused, deprecate after pytorch-translate migration
    def generate_batched_itr(self, data_itr, beam_size=None, cuda=False, timer=None):
        """Iterate over a batched dataset and yield individual translations.
        Args:
            cuda (bool, optional): use GPU for generation
            timer (StopwatchMeter, optional): time generations
        """
        for sample in data_itr:
            s = utils.move_to_cuda(sample) if cuda else sample
            if "net_input" not in s:
                continue
            input = s["net_input"]
            # model.forward normally channels prev_output_tokens into the decoder
            # separately, but SequenceGenerator directly calls model.encoder
            encoder_input = {
                k: v for k, v in input.items() if k != "prev_output_tokens"
            }
            if timer is not None:
                timer.start()
            with torch.no_grad():
                hypos = self.generate(encoder_input)
            if timer is not None:
                timer.stop(sum(len(h[0]["tokens"]) for h in hypos))
            for i, id in enumerate(s["id"].data):
                # remove padding
                src = utils.strip_pad(input["src_tokens"].data[i, :], self.pad)
                ref = (
                    utils.strip_pad(s["target"].data[i, :], self.pad)
                    if s["target"] is not None
                    else None
                )
                yield id, src, ref, hypos[i]

    @torch.no_grad()
    def generate(self, models, sample: Dict[str, Dict[str, Tensor]], **kwargs):
        """Generate translations. Match the api of other fairseq generators.

        Args:
            models (List[~fairseq.models.FairseqModel]): ensemble of models
            sample (dict): batch
            prefix_tokens (torch.LongTensor, optional): force decoder to begin
                with these tokens
            constraints (torch.LongTensor, optional): force decoder to include
                the list of constraints
            bos_token (int, optional): beginning of sentence token
                (default: self.eos)
        """
        return self._generate(sample, **kwargs)

    def _generate(
        self,
        sample: Dict[str, Dict[str, Tensor]],

        ## in sample  we need input sequence ids, 
        prefix_tokens: Optional[Tensor] = None,
        constraints: Optional[Tensor] = None,
        bos_token: Optional[int] = None,
        prefix_mention_is: bool = False,
    ):
        incremental_states = torch.jit.annotate(
            List[Dict[str, Dict[str, Optional[Tensor]]]],
            [
                torch.jit.annotate(Dict[str, Dict[str, Optional[Tensor]]], {})
                for i in range(self.model.models_size)
            ],
        )

        # TODO: create encoder input function, from sample to net_input
        net_input = sample["net_input"]

        if "input_ids" in net_input:
            src_tokens = net_input["input_ids"].cuda().to(self.model.single_model.device)
            # length of the source text being the character length except EndOfSentence and pad
            src_lengths = (
                (src_tokens.ne(self.eos) & src_tokens.ne(self.pad)).long().sum(dim=1)
            )
        # elif "source" in net_input:
        #     src_tokens = net_input["source"]
        #     src_lengths = (
        #         net_input["padding_mask"].size(-1) - net_input["padding_mask"].sum(-1)
        #         if net_input["padding_mask"] is not None
        #         else torch.tensor(src_tokens.size(-1)).to(src_tokens)
        #     )
        # else:
        #     raise Exception("expected src_tokens or source in net input")

        # bsz: total number of sentences in beam
        # Note that src_tokens may have more than 2 dimensions (i.e. audio features)
        bsz, src_len = src_tokens.size()[:2]
        beam_size = self.beam_size

        if constraints is not None and not self.search.supports_constraints:
            raise NotImplementedError(
                "Target-side constraints were provided, but search method doesn't support them"
            )

        # Initialize constraints, when active
        self.search.init_constraints(constraints, beam_size)
        max_len: int = -1
        if self.match_source_len:
            max_len = src_lengths.max().item()
        else:
            max_len = min(
                int(self.max_len_a * src_len + self.max_len_b),
                # exclude the EOS marker
                self.model.max_decoder_positions() - 1,
            )
        assert (
            self.min_len <= max_len
        ), "min_len cannot be larger than max_len, please adjust these!"
        # compute the encoder output for each beam

        ## TODO: modify encoder outputs,  :check
        encoder_outs = self.model.forward_encoder(net_input)

        # placeholder of indices for bsz * beam_size to hold tokens and accumulative scores
        new_order = torch.arange(bsz).view(-1, 1).repeat(1, beam_size).view(-1)
        new_order = new_order.to(src_tokens.device).long()

        ## TODO adjust following function  : check
        encoder_outs = self.model.reorder_encoder_out(encoder_outs, new_order)
        # ensure encoder_outs is a List.
        assert encoder_outs is not None

        # initialize buffers
        scores = (
            torch.zeros(bsz * beam_size, max_len + 1).to(src_tokens).float()
        )  # +1 for eos; pad is never chosen for scoring
        tokens = (
            torch.zeros(bsz * beam_size, max_len + 2)
            .to(src_tokens)
            .long()
            .fill_(self.pad)
        )  # +2 for eos and pad
        tokens[:, 0] = self.eos if bos_token is None else bos_token
        attn: Optional[Tensor] = None

        # A list that indicates candidates that should be ignored.
        # For example, suppose we're sampling and have already finalized 2/5
        # samples. Then cands_to_ignore would mark 2 positions as being ignored,
        # so that we only finalize the remaining 3 samples.
        cands_to_ignore = (
            torch.zeros(bsz, beam_size).to(src_tokens).eq(-1)
        )  # forward and backward-compatible False mask

        # list of completed sentences
        finalized = torch.jit.annotate(
            List[List[Dict[str, Tensor]]],
            [torch.jit.annotate(List[Dict[str, Tensor]], []) for i in range(bsz)],
        )  # contains lists of dictionaries of infomation about the hypothesis being finalized at each step

        finished = [
            False for i in range(bsz)
        ]  # a boolean array indicating if the sentence at the index is finished or not
        num_remaining_sent = bsz  # number of sentences remaining

        # number of candidate hypos per step
        cand_size = 2 * beam_size  # 2 x beam size in case half are EOS

        # offset arrays for converting between different indexing schemes
        bbsz_offsets = (
            (torch.arange(0, bsz) * beam_size)
            .unsqueeze(1)
            .type_as(tokens)
            .to(src_tokens.device)
        )
        cand_offsets = torch.arange(0, cand_size).type_as(tokens).to(src_tokens.device)

        reorder_state: Optional[Tensor] = None
        batch_idxs: Optional[Tensor] = None

        original_batch_idxs: Optional[Tensor] = None
        if "id" in sample and isinstance(sample["id"], Tensor):
            original_batch_idxs = sample["id"]
        else:
            original_batch_idxs = torch.arange(0, bsz).type_as(tokens)

        for step in range(max_len + 1):  # one extra step for EOS marker
            # reorder decoder internal states based on the prev choice of beams
            if reorder_state is not None:
                if batch_idxs is not None:
                    # update beam indices to take into account removed sentences
                    corr = batch_idxs - torch.arange(batch_idxs.numel()).type_as(
                        batch_idxs
                    )
                    reorder_state.view(-1, beam_size).add_(
                        corr.unsqueeze(-1) * beam_size
                    )
                    original_batch_idxs = original_batch_idxs[batch_idxs]

                ### TODO:modify function below: check, BART is not incremental
                # self.model.reorder_incremental_state(incremental_states, reorder_state)
                
                encoder_outs = self.model.reorder_encoder_out(
                    encoder_outs, reorder_state
                )

            ### TODO: modify decoder forward function : check, 
            ### lprobs is the outputs been through log_softmax

            lprobs, avg_attn_scores = self.model.forward_decoder(
                tokens[:, : step + 1],
                encoder_outs,
                incremental_states,
                self.temperature,
            )

            if self.lm_model is not None:
                lm_out = self.lm_model(tokens[:, : step + 1])
                probs = self.lm_model.get_normalized_probs(
                    lm_out, log_probs=True, sample=None
                )
                probs = probs[:, -1, :] * self.lm_weight
                lprobs += probs

            lprobs[lprobs != lprobs] = torch.tensor(-math.inf).to(lprobs)

            lprobs[:, self.pad] = -math.inf  # never select pad
            if self.unk is not None:
                lprobs[:, self.unk] -= self.unk_penalty  # apply unk penalty

            # handle max length constraint
            if step >= max_len:
                lprobs[:, : self.eos] = -math.inf
                lprobs[:, self.eos + 1 :] = -math.inf

            # handle prefix tokens (possibly with different lengths)
            if (
                prefix_tokens is not None
                and step < prefix_tokens.size(1)
                and step < max_len
            ):
                lprobs, tokens, scores = self._prefix_tokens(
                    step, lprobs, scores, tokens, prefix_tokens, beam_size
                )
            elif step < self.min_len:
                # minimum length constraint (does not apply if using prefix_tokens)
                lprobs[:, self.eos] = -math.inf

            # Record attention scores, only support avg_attn_scores is a Tensor
            if avg_attn_scores is not None:
                if attn is None:
                    attn = torch.empty(
                        bsz * beam_size, avg_attn_scores.size(1), max_len + 2
                    ).to(scores)
                attn[:, :, step + 1].copy_(avg_attn_scores)

            scores = scores.type_as(lprobs)
            eos_bbsz_idx = torch.empty(0).to(
                tokens
            )  # indices of hypothesis ending with eos (finished sentences)
            eos_scores = torch.empty(0).to(
                scores
            )  # scores of hypothesis ending with eos (finished sentences)



            if self.should_set_src_lengths:
                self.search.set_src_lengths(src_lengths)

            # if self.repeat_ngram_blocker is not None:
            #     lprobs = self.repeat_ngram_blocker(tokens, lprobs, bsz, beam_size, step)

            # Shape: (batch, cand_size)
            cand_scores, cand_indices, cand_beams = self.search.step(
                step,
                lprobs.view(bsz, -1, self.vocab_size),
                scores.view(bsz, beam_size, -1)[:, :, :step],
                tokens[:, : step + 1],
                original_batch_idxs,
                prefix_mention_is,
                prefix_tokens.size(1) if prefix_mention_is else None,
                step < prefix_tokens.size(1) if prefix_mention_is else None,
            )

            # cand_bbsz_idx contains beam indices for the top candidate
            # hypotheses, with a range of values: [0, bsz*beam_size),
            # and dimensions: [bsz, cand_size]
            cand_bbsz_idx = cand_beams.add(bbsz_offsets)

            # finalize hypotheses that end in eos
            # Shape of eos_mask: (batch size, beam size)
            eos_mask = cand_indices.eq(self.eos) & cand_scores.ne(-math.inf)
            eos_mask[:, :beam_size][cands_to_ignore] = torch.tensor(0).to(eos_mask)

            # only consider eos when it's among the top beam_size indices
            # Now we know what beam item(s) to finish
            # Shape: 1d list of absolute-numbered
            eos_bbsz_idx = torch.masked_select(
                cand_bbsz_idx[:, :beam_size], mask=eos_mask[:, :beam_size]
            )

            finalized_sents: List[int] = []
            if eos_bbsz_idx.numel() > 0:
                eos_scores = torch.masked_select(
                    cand_scores[:, :beam_size], mask=eos_mask[:, :beam_size]
                )

                finalized_sents = self.finalize_hypos(
                    step,
                    eos_bbsz_idx,
                    eos_scores,
                    tokens,
                    scores,
                    finalized,
                    finished,
                    beam_size,
                    attn,
                    src_lengths,
                    max_len,
                    prefix_mention_is,
                    prefix_tokens.size(1) if prefix_mention_is else None,
                )
                num_remaining_sent -= len(finalized_sents)

            assert num_remaining_sent >= 0
            if num_remaining_sent == 0:
                break
            if self.search.stop_on_max_len and step >= max_len:
                break
            assert step < max_len, f"{step} < {max_len}"

            # Remove finalized sentences (ones for which {beam_size}
            # finished hypotheses have been generated) from the batch.
            if len(finalized_sents) > 0:
                new_bsz = bsz - len(finalized_sents)

                # construct batch_idxs which holds indices of batches to keep for the next pass
                batch_mask = torch.ones(
                    bsz, dtype=torch.bool, device=cand_indices.device
                )
                batch_mask[finalized_sents] = False
                # TODO replace `nonzero(as_tuple=False)` after TorchScript supports it
                batch_idxs = torch.arange(
                    bsz, device=cand_indices.device
                ).masked_select(batch_mask)

                # Choose the subset of the hypothesized constraints that will continue
                self.search.prune_sentences(batch_idxs)

                eos_mask = eos_mask[batch_idxs]
                cand_beams = cand_beams[batch_idxs]
                bbsz_offsets.resize_(new_bsz, 1)
                cand_bbsz_idx = cand_beams.add(bbsz_offsets)
                cand_scores = cand_scores[batch_idxs]
                cand_indices = cand_indices[batch_idxs]

                if prefix_tokens is not None:
                    prefix_tokens = prefix_tokens[batch_idxs]
                src_lengths = src_lengths[batch_idxs]
                cands_to_ignore = cands_to_ignore[batch_idxs]

                scores = scores.view(bsz, -1)[batch_idxs].view(new_bsz * beam_size, -1)
                tokens = tokens.view(bsz, -1)[batch_idxs].view(new_bsz * beam_size, -1)
                if attn is not None:
                    attn = attn.view(bsz, -1)[batch_idxs].view(
                        new_bsz * beam_size, attn.size(1), -1
                    )
                bsz = new_bsz
            else:
                batch_idxs = None

            # Set active_mask so that values > cand_size indicate eos hypos
            # and values < cand_size indicate candidate active hypos.
            # After, the min values per row are the top candidate active hypos

            # Rewrite the operator since the element wise or is not supported in torchscript.

            eos_mask[:, :beam_size] = ~((~cands_to_ignore) & (~eos_mask[:, :beam_size]))
            active_mask = torch.add(
                eos_mask.type_as(cand_offsets) * cand_size,
                cand_offsets[: eos_mask.size(1)],
            )

            # get the top beam_size active hypotheses, which are just
            # the hypos with the smallest values in active_mask.
            # {active_hypos} indicates which {beam_size} hypotheses
            # from the list of {2 * beam_size} candidates were
            # selected. Shapes: (batch size, beam size)
            new_cands_to_ignore, active_hypos = torch.topk(
                active_mask, k=beam_size, dim=1, largest=False
            )

            # update cands_to_ignore to ignore any finalized hypos.
            cands_to_ignore = new_cands_to_ignore.ge(cand_size)[:, :beam_size]
            # Make sure there is at least one active item for each sentence in the batch.
            assert (~cands_to_ignore).any(dim=1).all()

            # update cands_to_ignore to ignore any finalized hypos

            # {active_bbsz_idx} denotes which beam number is continued for each new hypothesis (a beam
            # can be selected more than once).
            active_bbsz_idx = torch.gather(cand_bbsz_idx, dim=1, index=active_hypos)
            active_scores = torch.gather(cand_scores, dim=1, index=active_hypos)

            active_bbsz_idx = active_bbsz_idx.view(-1)
            active_scores = active_scores.view(-1)

            # copy tokens and scores for active hypotheses

            # Set the tokens for each beam (can select the same row more than once)
            tokens[:, : step + 1] = torch.index_select(
                tokens[:, : step + 1], dim=0, index=active_bbsz_idx
            )
            # Select the next token for each of them
            tokens.view(bsz, beam_size, -1)[:, :, step + 1] = torch.gather(
                cand_indices, dim=1, index=active_hypos
            )
            if step > 0:
                scores[:, :step] = torch.index_select(
                    scores[:, :step], dim=0, index=active_bbsz_idx
                )
            scores.view(bsz, beam_size, -1)[:, :, step] = torch.gather(
                cand_scores, dim=1, index=active_hypos
            )

            # Update constraints based on which candidates were selected for the next beam
            self.search.update_constraints(active_hypos)

            # copy attention for active hypotheses
            if attn is not None:
                attn[:, :, : step + 2] = torch.index_select(
                    attn[:, :, : step + 2], dim=0, index=active_bbsz_idx
                )

            # reorder incremental state in decoder
            reorder_state = active_bbsz_idx

        # sort by score descending
        for sent in range(len(finalized)):
            scores = torch.tensor(
                [float(elem["score"].item()) for elem in finalized[sent]]
            )
            _, sorted_scores_indices = torch.sort(scores, descending=True)
            finalized[sent] = [finalized[sent][ssi] for ssi in sorted_scores_indices]
            finalized[sent] = torch.jit.annotate(
                List[Dict[str, Tensor]], finalized[sent]
            )


        ### TODO: change the output form: check
        results = []
        scores = []
        for batch in range(len(finalized)):
            result = []
            score = []
            for beam in range(len(finalized[batch])):
                result.append(finalized[batch][beam]['tokens'])
                score.append(finalized[batch][beam]['positional_scores'])
            results.append(result)
            scores.append(score)

        return results, scores

    def _prefix_tokens(
        self, step: int, lprobs, scores, tokens, prefix_tokens, beam_size: int
    ):
        """Handle prefix tokens"""
        prefix_toks = prefix_tokens[:, step].unsqueeze(-1).repeat(1, beam_size).view(-1)
        prefix_lprobs = lprobs.gather(-1, prefix_toks.unsqueeze(-1))
        prefix_mask = prefix_toks.ne(self.pad)
        lprobs[prefix_mask] = torch.tensor(-math.inf).to(lprobs)
        lprobs[prefix_mask] = lprobs[prefix_mask].scatter(
            -1, prefix_toks[prefix_mask].unsqueeze(-1), prefix_lprobs[prefix_mask]
        )
        # if prefix includes eos, then we should make sure tokens and
        # scores are the same across all beams
        eos_mask = prefix_toks.eq(self.eos)
        if eos_mask.any():
            # validate that the first beam matches the prefix
            first_beam = tokens[eos_mask].view(-1, beam_size, tokens.size(-1))[
                :, 0, 1 : step + 1
            ]
            eos_mask_batch_dim = eos_mask.view(-1, beam_size)[:, 0]
            target_prefix = prefix_tokens[eos_mask_batch_dim][:, :step]
            assert (first_beam == target_prefix).all()

            # copy tokens, scores and lprobs from the first beam to all beams
            tokens = self.replicate_first_beam(tokens, eos_mask_batch_dim, beam_size)
            scores = self.replicate_first_beam(scores, eos_mask_batch_dim, beam_size)
            lprobs = self.replicate_first_beam(lprobs, eos_mask_batch_dim, beam_size)
        return lprobs, tokens, scores

    def replicate_first_beam(self, tensor, mask, beam_size: int):
        tensor = tensor.view(-1, beam_size, tensor.size(-1))
        tensor[mask] = tensor[mask][:, :1, :]
        return tensor.view(-1, tensor.size(-1))

    def finalize_hypos(
        self,
        step: int,
        bbsz_idx,
        eos_scores,
        tokens,
        scores,
        finalized: List[List[Dict[str, Tensor]]],
        finished: List[bool],
        beam_size: int,
        attn: Optional[Tensor],
        src_lengths,
        max_len: int,
        prefix_mention_is = False,
        prefix_token_num = None,
    ):
        """Finalize hypothesis, store finalized information in `finalized`, and change `finished` accordingly.
        A sentence is finalized when {beam_size} finished items have been collected for it.

        Returns number of sentences (not beam items) being finalized.
        These will be removed from the batch and not processed further.
        Args:
            bbsz_idx (Tensor):
        """
        assert bbsz_idx.numel() == eos_scores.numel()

        # clone relevant token and attention tensors.
        # tokens is (batch * beam, max_len). So the index_select
        # gets the newly EOS rows, then selects cols 1..{step + 2}
        tokens_clone = tokens.index_select(0, bbsz_idx)[
            :, 1 : step + 2
        ]  # skip the first index, which is EOS

        tokens_clone[:, step] = self.eos
        attn_clone = (
            attn.index_select(0, bbsz_idx)[:, :, 1 : step + 2]
            if attn is not None
            else None
        )

        # compute scores per token position
        pos_scores = scores.index_select(0, bbsz_idx)[:, : step + 1]
        pos_scores[:, step] = eos_scores
        # convert from cumulative to per-position scores
        pos_scores[:, 1:] = pos_scores[:, 1:] - pos_scores[:, :-1]

        # normalize sentence-level scores
        if self.normalize_scores:
            eos_scores /= (step + 1) ** self.len_penalty

        # cum_unfin records which sentences in the batch are finished.
        # It helps match indexing between (a) the original sentences
        # in the batch and (b) the current, possibly-reduced set of
        # sentences.
        cum_unfin: List[int] = []
        prev = 0
        for f in finished:
            if f:
                prev += 1
            else:
                cum_unfin.append(prev)

        # The keys here are of the form "{sent}_{unfin_idx}", where
        # "unfin_idx" is the index in the current (possibly reduced)
        # list of sentences, and "sent" is the index in the original,
        # unreduced batch
        # set() is not supported in script export
        sents_seen: Dict[str, Optional[Tensor]] = {}

        # For every finished beam item
        for i in range(bbsz_idx.size()[0]):
            idx = bbsz_idx[i]
            score = eos_scores[i]
            # sentence index in the current (possibly reduced) batch
            unfin_idx = idx // beam_size
            # sentence index in the original (unreduced) batch
            sent = unfin_idx + cum_unfin[unfin_idx]
            # Cannot create dict for key type '(int, int)' in torchscript.
            # The workaround is to cast int to string
            seen = str(sent.item()) + "_" + str(unfin_idx.item())
            if seen not in sents_seen:
                sents_seen[seen] = None

            if self.match_source_len and step > src_lengths[unfin_idx]:
                score = torch.tensor(-math.inf).to(score)

            # An input sentence (among those in a batch) is finished when
            # beam_size hypotheses have been collected for it
            if len(finalized[sent]) < beam_size:
                if attn_clone is not None:
                    # remove padding tokens from attn scores
                    hypo_attn = attn_clone[i]
                else:
                    hypo_attn = torch.empty(0)
                
                if prefix_mention_is:
                    finalized[sent].append(
                        {
                            "tokens": tokens_clone[i],
                            "score": torch.sum(pos_scores[i][list(tokens_clone[i]).index(16)+1:]) / (len(pos_scores[i])-list(tokens_clone[i]).index(16)-1)**self.len_penalty,
                            # "score": torch.sum(pos_scores[i][prefix_token_num+1:]) / (len(pos_scores[i])-prefix_token_num-1)**self.len_penalty,
                            "attention": hypo_attn,  # src_len x tgt_len
                            "alignment": torch.empty(0),
                            # "positional_scores": pos_scores[i][prefix_token_num+1:],
                            "positional_scores": pos_scores[i][list(tokens_clone[i]).index(16)+1:],
                        }
                    )
                else:
                    finalized[sent].append(
                        {
                            "tokens": tokens_clone[i],
                            "score": score,
                            "attention": hypo_attn,  # src_len x tgt_len
                            "alignment": torch.empty(0),
                            "positional_scores": pos_scores[i],
                        }
                    )

        newly_finished: List[int] = []

        for seen in sents_seen.keys():
            # check termination conditions for this sentence
            sent: int = int(float(seen.split("_")[0]))
            unfin_idx: int = int(float(seen.split("_")[1]))

            if not finished[sent] and self.is_finished(
                step, unfin_idx, max_len, len(finalized[sent]), beam_size
            ):
                finished[sent] = True
                newly_finished.append(unfin_idx)

        return newly_finished

    def is_finished(
        self,
        step: int,
        unfin_idx: int,
        max_len: int,
        finalized_sent_len: int,
        beam_size: int,
    ):
        """
        Check whether decoding for a sentence is finished, which
        occurs when the list of finalized sentences has reached the
        beam size, or when we reach the maximum length.
        """
        assert finalized_sent_len <= beam_size
        if finalized_sent_len == beam_size or step == max_len:
            return True
        return False


class EnsembleModel(nn.Module):
    """A wrapper around an ensemble of models."""

    def __init__(self, models):
        super().__init__()
        self.models_size = 1 #len(models) I think this means number of models

        self.single_model = models
        self.models = nn.ModuleList([models])

        self.has_incremental: bool = self.single_model.config.is_encoder_decoder

    def forward(self):
        pass

    def has_encoder(self):
        return self.single_model.config.is_encoder_decoder

    def has_incremental_states(self):
        return self.has_incremental

    def max_decoder_positions(self):
        return min([m.config.max_length for m in self.models])

    @torch.jit.export
    def forward_encoder(self, net_input: Dict[str, Tensor]):

        encoder = self.single_model.get_encoder()
        # encoder_kwargs = {
        #     argument: value
        #     for argument, value in net_input.items()
        #     if not (argument.startswith("decoder_") or argument.startswith("cross_attn") or argument == "inputs_embeds" or argument == "attention_mask")
        # }
        ###### add 'or argument == "inputs_embeds" or argument == "attention_mask"' because we modify this below

        ###### add encoder input ids to soft prompt + embedding input vectors along with corresponding attention mask
        input_ids = net_input['input_ids'].cuda().to(self.single_model.device)
        if 'mention_index' in net_input:
            mention_indexes = net_input['mention_index']
        else:
            mention_indexes = None
        inputs_embeds_enc = self.single_model.id2emb_cat_prompt(input_ids, coder_part = 'encoder', insert_idx = mention_indexes)

        attention_mask = net_input['attention_mask'].cuda().to(self.single_model.device)
        if attention_mask is not None:
            attention_mask_enc = self.single_model.extend_attention_mask(attention_mask, self.single_model.n_tokens_enc)
        else:
            attention_mask_enc = None
        # if self.n_tokens_enc > 0:
        #     attention_mask_enc = torch.cat([torch.full((attention_mask_enc.shape[0], self.n_tokens_enc), 1).to(self.device), attention_mask_enc],dim=1)
        ######  add part end

        encoder_outputs = encoder(input_ids = None, inputs_embeds = inputs_embeds_enc, attention_mask = attention_mask_enc, return_dict=True)#, **encoder_kwargs)
        ###### change input_ids input to None, add inputs_embeds and attention_mask variables.


        
        return [encoder_outputs]

    @torch.jit.export
    def forward_decoder(
        self,
        tokens,
        encoder_outs: List[Dict[str, List[Tensor]]],
        incremental_states: List[Dict[str, Dict[str, Optional[Tensor]]]],
        temperature: float = 1.0,
    ):
        log_probs = []
        # avg_attn: Optional[Tensor] = None
        # encoder_out: Optional[Dict[str, List[Tensor]]] = None
        for i, model in enumerate(self.models):
            if self.has_encoder():
                encoder_out = encoder_outs[i]
            # decode each model

            ### not incremental
            # if self.has_incremental_states():
            #     decoder_out = model.decoder.forward(
            #         tokens,
            #         encoder_out=encoder_out,
            #         incremental_state=incremental_states[i],
            #     )
            # else:
            decoder_out = model(
                decoder_input_ids = tokens,
                encoder_outputs = encoder_out, 
                return_dict=True,
                output_attentions=False,
                output_hidden_states=False,
            )


            attn: Optional[Tensor] = None
            ### no need for attn
            # decoder_len = len(decoder_out)
            decoder_len = 0
            if decoder_len > 1 and decoder_out[1] is not None:
                if isinstance(decoder_out[1], Tensor):
                    attn = decoder_out[1]
                else:
                    attn_holder = decoder_out[1]["attn"]
                    if isinstance(attn_holder, Tensor):
                        attn = attn_holder
                    elif attn_holder is not None:
                        attn = attn_holder[0]
                if attn is not None:
                    attn = attn[:, -1, :]

            decoder_out_tuple = (decoder_out.logits[:, -1:, :], None)

            probs = F.log_softmax(decoder_out_tuple[0], dim=-1, dtype=torch.float32)

            probs = probs[:, -1, :]
            if self.models_size == 1:
                return probs, attn

            log_probs.append(probs)
            if attn is not None:
                if avg_attn is None:
                    avg_attn = attn
                else:
                    avg_attn.add_(attn)

        avg_probs = torch.logsumexp(torch.stack(log_probs, dim=0), dim=0) - math.log(
            self.models_size
        )

        if avg_attn is not None:
            avg_attn.div_(self.models_size)
        return avg_probs, avg_attn

    @torch.jit.export
    def reorder_encoder_out(
        self, encoder_outs: Optional[List[Dict[str, List[Tensor]]]], new_order
    ):  
        for i in range(len(encoder_outs)):
            encoder_outs[i]['last_hidden_state'] = encoder_outs[i]['last_hidden_state'].index_select(0, new_order)

                
        return encoder_outs

    @torch.jit.export
    def reorder_incremental_state(
        self,
        incremental_states: List[Dict[str, Dict[str, Optional[Tensor]]]],
        new_order,
    ):
        if not self.has_incremental_states():
            return
        for i, model in enumerate(self.models):
            model.decoder.reorder_incremental_state_scripting(
                incremental_states[i], new_order
            )





import spacy
from bioel.ontology import BiomedicalOntology
from bioel.models.scispacy.candidate_generation import CandidateGenerator
from bioel.models.scispacy.scispacy_embeddings import KnowledgeBaseEmbeddings
from bioel.models.scispacy.entity_linking import EntityLinker
from bioel.utils.bigbio_utils import (
    load_bigbio_dataset,
    add_deabbreviations,
    dataset_to_documents,
    dataset_to_df,
    load_dataset_df,
    resolve_abbreviation,
    CUIS_TO_REMAP,
    CUIS_TO_EXCLUDE,
    DATASET_NAMES,
    VALIDATION_DOCUMENT_IDS,
)
from bioel.utils.dataset_consts import (
    dataset_to_pretty_name,
    model_to_pretty_name,
    model_to_color,
)
import ujson
import os
import pickle


def evaluate_model(
    params,
    model,
):
    """
    Params :
    ---------
    dataset: str
        Name of the dataset to evaluate
    ontology: BiomedicalOntology
        Ontology to use for candidate generation
    k: int
        Number of candidates to generate
    path_to_save: str
        Path to save the serialized knowledge base
    output_path: str
        Path to save the output
    equivalent_cuis: bool
        Whether the ontology has equivalent cuis or not
    path_to_abbrev: str
        Path to the abbreviations file
    """
    if hasattr(BiomedicalOntology, params["load_function"]):
        load_func = getattr(BiomedicalOntology, params["load_function"])
        if params["ontology_dict"]:
            ontology_object = load_func(**params["ontology_dict"])
            print(f"Ontology loaded successfully. Name: {ontology_object.name}")
        else:
            raise ValueError("No ontology data provided.")
    else:
        raise ValueError(
            f"Error: {params['load_function']} is not a valid function for BiomedicalOntology."
        )

    myembed = KnowledgeBaseEmbeddings(ontology_object)
    myembed.create_tfidf_ann_index(params["path_to_save"])
    kb = myembed.serialized_kb()
    cand_gen = CandidateGenerator(kb)
    df = load_dataset_df(
        name=params["dataset"], path_to_abbrev=params["path_to_abbrev"]
    )
    df = df.query("split == 'test'")

    output = list()

    if params["equivalant_cuis"]:
        cui_synsets = {}
        for cui, entity in ontology_object.entities.items():
            cui_synsets[cui] = entity.equivalant_cuis

    for index, row in df.iterrows():
        list_dbid = list(row.db_ids)  # Accessing db_ids as a list
        if params["path_to_abbrev"]:
            mention = row.deabbreviated_text  # Accessing the text of the current row
        else:
            mention = row.text
        candidates = cand_gen([mention], 3 * params["k"])  # Generating candidates
        predicted = []

        for cand in candidates[0]:
            score = max(cand.similarities)
            predicted.append((cand.concept_id, score))

        sorted_predicted = sorted(predicted, reverse=True, key=lambda x: x[1])
        sorted_predicted = sorted_predicted[: params["k"]]  # Taking top k predictions

        cui_result = [[cui[0]] for cui in sorted_predicted]
        score = [[cui[1]] for cui in sorted_predicted]

        if params["equivalant_cuis"]:
            cui_result = [cui_synsets[y[0]] for y in cui_result]

        if params["path_to_abbrev"]:
            output.append(
                {
                    "document_id": row.document_id,
                    "offsets": row.offsets,
                    "text": row.text,
                    "type": row.type,
                    "db_ids": list_dbid,
                    "split": row.split,
                    "deabbreviated_text": row.deabbreviated_text,
                    "mention_id": row.mention_id + ".abbr_resolved",
                    "candidates": cui_result,
                    "scores": score,
                }
            )
        else:
            output.append(
                {
                    "document_id": row.document_id,
                    "offsets": row.offsets,
                    "text": row.text,
                    "type": row.type,
                    "db_ids": list_dbid,
                    "split": row.split,
                    "mention_id": row.mention_id,
                    "candidates": cui_result,
                    "scores": score,
                }
            )

    with open(params["output_path"], "w") as f:
        f.write(ujson.dumps(output, indent=2))



import logging
import sys
import ujson


from bioel.models.sapbert.utils import (
    sapbert_predict,
    parse_args_bigbio,
    init_logging,
    load_dictionary,
    make_unique_model_savepath,
)
from torch.utils.data import DataLoader


import wandb

from bioel.models.sapbert.data.utils import SapBertBigBioDataset
from bioel.ontology import BiomedicalOntology
from bioel.models.sapbert.model.model_wrapper import Model_Wrapper


LOGGER = logging.getLogger()


def sapbert_collate_fn(batch):
    mentions = [x["text"] for x in batch]
    # labels = [x["cuis"] for x in batch]
    labels = [x["db_ids"] for x in batch]
    metadata = batch

    return mentions, labels, metadata


def evaluate_model(config, model_wrapper):

    # Get the Alias from the Big Bio dataset into a dictionary
    eval_dictionary = load_dictionary(dictionary_path=config["train_dir"])

    if config["path_to_abbreviation"] is None:
        resolve_abbreviations = False
    else:
        resolve_abbreviations = True

    # Load data
    data = SapBertBigBioDataset(
        config["dataset_name"],
        splits_to_include=[config["split"]],
        path_to_abbreviation_dict=config["path_to_abbreviation"],
        resolve_abbreviations=resolve_abbreviations,
    )
    loader = DataLoader(
        data, collate_fn=sapbert_collate_fn, batch_size=config["eval_batch_size"]
    )

    # Load Model
    model = model_wrapper.load_model(
        path=config["model_dir"],
        max_length=config["max_length"],
        use_cuda=config["use_cuda"],
    )

    # Get unique save path
    preds_save_path = make_unique_model_savepath(
        dir=config["output_dir"],
        model="sapbert",
        dataset=config["dataset_name"],
        file_ext="json",
    )

    dict_cache_filepath = config["dict_cache_path"]
    LOGGER.info(f"Dict CACHE Path: {dict_cache_filepath}")
    LOGGER.info(f"Debugging mode? {config['debug']}")

    results = sapbert_predict(
        model_wrapper=model,
        eval_dictionary=eval_dictionary,
        dataloader=loader,
        batch_size=config["eval_batch_size"],
        topk=config["topk"],
        dict_cache_filepath=dict_cache_filepath,
        debug=config["debug"],
    )

    if resolve_abbreviations:
        for element in results:
            element["mention_id"] = element["mention_id"] + ".abbr_resolved"

    LOGGER.info(f"Saving results to {preds_save_path}")
    with open(preds_save_path, "w") as f:
        f.write(ujson.dumps(results, indent=2))


# if __name__ == "__main__":
#     init_logging()
#     args = parse_args_bigbio()
#     main(args)



from bioel.models.biogenel.trie import Trie
from bioel.models.biogenel.utils import reform_input
from bioel.models.biogenel.LightningDataModule import BioGenELDataModule
from bioel.models.biogenel.LightningModule import BioGenElLightningModule
from bioel.ontology import BiomedicalOntology
from transformers import BartTokenizer

def evaluate_model(params, model: BioGenElLightningModule):
    tokenizer = BartTokenizer.from_pretrained(f"{params.model_token_path}")
    if hasattr(BiomedicalOntology, params.load_function):
        load_func = getattr(BiomedicalOntology, params.load_function)
        if params.ontology_dict:
            ontology_object = load_func(**params.ontology_dict)
            print(f"Ontology loaded successfully. Name: {ontology_object.name}")
        else:
            raise ValueError("No ontology data provided.")
    else:
        raise ValueError(
            f"Error: {params.load_function} is not a valid function for BiomedicalOntology."
        )
    datamodule = BioGenELDataModule(tokenizer, 
                    save_data_dir=f"{params.dataset_path}/", 
                    dataset_name = params.dataset_name, 
                    ontology = ontology_object,
                    prefix_mention_is= params.prefix_mention_is,
                    evaluate= params.evaluation,
                    resolve_abbrevs= params.resolve_abbrevs,
                    preprocess_data = params.preprocess_data,
                    path_to_abbrev = params.path_to_abbrev,
                    batch_size = 1)
    
    model.init_datamodule(datamodule=datamodule)
    model.evaluate()


# Copyright (c) Facebook, Inc. and its affiliates.
# Copyright (c) 2021 Dhruv Agarwal and authors of arboEL.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

import sys
import os
import torch
import torch.nn.functional as F

from transformers import AutoModel
from bioel.models.arboel.biencoder.model.common.ranker_base import (
    BertEncoder,
    get_model_obj,
)
from bioel.models.arboel.biencoder.model.common.optimizer import get_bert_optimizer

from IPython import embed


def load_biencoder(params):
    # Init model
    biencoder = BiEncoderRanker(params)
    return biencoder


class BiEncoderModule(torch.nn.Module):
    def __init__(self, params):
        """
        Params:
        - params : dict
        Dictionary containing configuration options for the bi-encoder module
        """
        super(BiEncoderModule, self).__init__()
        ctxt_bert = AutoModel.from_pretrained(
            params["model_name_or_path"], return_dict=False
        )  # Could be a path containing config.json and pytorch_model.bin; or could be an id shorthand for a model that is loaded in the library
        cand_bert = AutoModel.from_pretrained(
            params["model_name_or_path"], return_dict=False
        )
        self.context_encoder = BertEncoder(
            ctxt_bert,
            params["out_dim"],
            layer_pulled=params[
                "pull_from_layer"
            ],  # from which layer shall we pull the encoded embedding
            add_linear=params[
                "add_linear"
            ],  # whether an additional linear transformation is applied to the output
        )
        self.cand_encoder = BertEncoder(
            cand_bert,
            params["out_dim"],
            layer_pulled=params["pull_from_layer"],
            add_linear=params["add_linear"],
        )
        self.config = (
            ctxt_bert.config
        )  # Configuration object contains all settings used to initialize the BERT model (number of layers, hidden units, attention heads, etc)

    def forward(
        self,
        token_idx_ctxt,
        segment_idx_ctxt,
        mask_ctxt,
        token_idx_cands,
        segment_idx_cands,
        mask_cands,
    ):
        """
        Returns a tuple containing two elements:
        embedding_ctxt (Tensor): The ENCODED embeddings for the context input. (None if token_idx_ctxt is None).
        embedding_cands (Tensor): The ENCODED embeddings for the candidate inputs. (None if token_idx_cands is None).
        ------
        Params :
        - token_idx_ctxt : Tensor
        Contains token indices for the context input. These indices correspond to the tokens positions in the BERT vocabulary.
        - segment_idx_ctxt : Tensor
        Tensor of segment indices for the context input, used for distinguishing different segments within the same input
        - mask_ctxt : Tensor
        Attention mask for the context input
        - token_idx_cands : Tensor
        Tensor of token indices for the candidate inputs.
        - segment_idx_cands : Tensor
        Tensor of segment indices for the candidate inputs.
        - mask_cands : Tensor
        Attention mask for the candidate inputs.
        """
        embedding_ctxt = None
        if token_idx_ctxt is not None:
            embedding_ctxt = self.context_encoder(
                token_idx_ctxt, segment_idx_ctxt, mask_ctxt
            )
        embedding_cands = None
        if token_idx_cands is not None:
            embedding_cands = self.cand_encoder(
                token_idx_cands, segment_idx_cands, mask_cands
            )
        return embedding_ctxt, embedding_cands


class BiEncoderRanker(torch.nn.Module):
    "Goal : Encode pieces of text into vector embeddings and then compare these embeddings to perform ranking"

    def __init__(self, params, shared=None):
        # Initialize an instance of the class with specific configurations, resources, and model components.
        super(BiEncoderRanker, self).__init__()
        self.params = params

        # init tokenizer
        self.NULL_IDX = 0
        self.START_TOKEN = "[CLS]"
        self.END_TOKEN = "[SEP]"

        # init bi-encoder model and stores it in self.model
        self.build_model()

        # model_path = params.get("path_to_biencoder_model", params.get("path_to_model"))
        # if model_path:
        #     self.load_model(model_path)

    def load_model(self, fname, cpu=False):
        """
        Loads pre-trained weights into the model from a file. If cpu is true, forces the model to load onto CPU memory.
        ------
        Params :
        - fname : str
        The file name (or path) where the model's state dictionary is saved.
        """
        # Load the entire checkpoint
        checkpoint = torch.load(fname, map_location="cpu" if cpu else None)

        # Access the model state_dict, stored under 'state_dict' key in PyTorch Lightning checkpoints
        state_dict = checkpoint["state_dict"]

        # Remove the 'model.' prefix that PyTorch Lightning adds
        new_state_dict = {
            key.replace("model.", ""): value for key, value in state_dict.items()
        }

        # Load the weights
        self.model.load_state_dict(new_state_dict)

    def build_model(self):
        """
        Constructs the bi-encoder model and stores it in self.model
        """
        self.model = BiEncoderModule(self.params)

    def load_model(self, fname, cpu=False):
        if cpu:
            state_dict = torch.load(fname, map_location=lambda storage, location: "cpu")
        else:
            state_dict = torch.load(fname)
        self.model.load_state_dict(state_dict)

    def get_optimizer(self, optim_states=None, saved_optim_type=None):
        """
        Initializes and returns an optimizer for the model
        ------
        Params:
        - optim_states : dict
        Contains the state information from a previously used optimizer (number of update, moments of gradient etc...)
        - saved_optim_type : str
        Contains the type of optimization
        """
        return get_bert_optimizer(
            [self.model],
            self.params["type_optimization"],  # Ex : "Adam", "SGD",
            self.params["learning_rate"],
            fp16=self.params.get("fp16"),
        )

    def encode_context(self, ctxt, requires_grad=False):
        """
        Process and encode contextual input with SELF.MODEL()
        ------
        Params:
        - ctxt : Tensor of dim (batch_size, sequence_length) : element (i,j) = INDICE of the j-ième token of the i-ème sequence in the batch
        Collection of token INDICES representing the context inputs
        - requires_grad : bool
        A flag indicating whether the resulting embeddings should require gradient computation
        """
        token_idx_ctxt, segment_idx_ctxt, mask_ctxt = to_bert_input(ctxt, self.NULL_IDX)
        embedding_context, _ = self.model(
            token_idx_ctxt, segment_idx_ctxt, mask_ctxt, None, None, None
        )
        """
        if not requires_grad:
            embedding_context = embedding_context.detach()  # Detach but keep on the same device
        return embedding_context
    
        """
        if requires_grad:
            return embedding_context
        return embedding_context.detach().cpu()

    def encode_candidate(self, cands, requires_grad=False):
        """
        Process and encode candidate input with SELF.MODEL()
        ------
        Params:
        - cands : Tensor of dim (cands_number, sequence_length) : element (i,j) = INDICE of the j-ième token of the i-ème sequence in the list of candidate
        Collection of token INDICES representing the candidate inputs
        - requires_grad : bool
        A flag indicating whether the resulting embeddings should require gradient computation
        """

        token_idx_cands, segment_idx_cands, mask_cands = to_bert_input(
            cands, self.NULL_IDX
        )

        _, embedding_cands = self.model(
            None, None, None, token_idx_cands, segment_idx_cands, mask_cands
        )  # dim = (cands_number, embedding_size)

        if requires_grad:
            return embedding_cands

        return embedding_cands.detach().cpu()

    # Score candidates given context input and label input
    # If cand_encs is provided (pre-computed), cand_vecs is ignored
    def score_candidate(
        self,
        text_vecs,
        cand_vecs,
        random_negs=True,
        cand_encs=None,  # pre-computed candidate encoding.
    ):
        """
        Designed to calculate the similarity scores between context inputs and candidate inputs.
        It supports both scenarios where candidate encodings are pre-computed and where they need to be computed on-the-fly.
        ------
        Params:
        - text_vecs : Tensor of dim (batch_size, sequence_length)
        Collection of token INDICES representing the context inputs
        - cands : Tensor of dim (n, sequence_length) where n is the number of candidate encodings
        Collection of token INDICES representing the candidate inputs
        - random_negs : bool
        A flag indicating whether to use random negatives for scoring
        - cand_encs : A tensor of dim (batch, embed_size)
        Ready for direct comparison with context embeddings
        ------
        - embedding_ctxt : Tensor of dim (batch_size, embed_size)
        - embedding_cands : Tensor of dim (n, embed_size) and Tensor of dim (batch_size, top_k, embed_size) where top_k is the number of candidates for each context
        """
        # Encode contexts first
        token_idx_ctxt, segment_idx_ctxt, mask_ctxt = to_bert_input(
            text_vecs, self.NULL_IDX
        )
        embedding_ctxt, _ = self.model(
            token_idx_ctxt,
            segment_idx_ctxt,
            mask_ctxt,
            None,
            None,
            None,  # None = token_idx_cands / segment_idx_cands / mask_cands
        )

        # Candidate encoding is given, do not need to re-compute
        # Directly return the score of context encoding and candidate encoding
        if cand_encs is not None:
            return embedding_ctxt.mm(cand_encs.t())

        # Train time. We compare with all elements of the batch
        token_idx_cands, segment_idx_cands, mask_cands = to_bert_input(
            cand_vecs, self.NULL_IDX
        )
        _, embedding_cands = self.model(
            None,
            None,
            None,
            token_idx_cands,
            segment_idx_cands,
            mask_cands,  # None = token_idx_ctxt / segment_idx_ctxt / mask_ctxt
        )
        if embedding_cands.shape[0] != embedding_ctxt.shape[0]:
            embedding_cands = embedding_cands.view(
                embedding_ctxt.shape[0],
                embedding_cands.shape[0] // embedding_ctxt.shape[0],
                embedding_cands.shape[1],
            )  # batchsize x topk x embed_size

        if random_negs:
            # train on random negatives
            ####### Issue here : .mm() expects both operands to be 2D matrices and embedding_cands is a 3D tensor
            ####### Not used because we always train on hard negatives
            return embedding_ctxt.mm(embedding_cands.t())
        else:
            # train on hard negatives
            embedding_ctxt = embedding_ctxt.unsqueeze(2)  # batchsize x embed_size x 1
            scores = torch.bmm(
                embedding_cands, embedding_ctxt
            )  # batchsize x topk x 1    ## Score = weight of edges
            scores = torch.squeeze(
                scores, dim=2
            )  # batchsize x topk   # Removes the dimension at index 2 of the scores tensor
            return scores  # element (i,j) = weight between mention_i and candidate_j

    # label_input -- negatives provided
    # If label_input is None, train on in-batch negatives
    def forward(
        self,
        context_input,
        cand_input=None,
        label_input=None,
        mst_data=None,
        pos_neg_loss=False,
        only_logits=False,
    ):
        """
        Computes the loss and score of the model for a batch of data

        Parameters
        ----------
        - context_input : Tensor of dim (batch_size, sequence_length)
        Token INDICES representing the context inputs
        - cand_input : Tensor of dim (batch_size, sequence_length)
        Token INDICES representing the candidate inputs
        - label_input : Tensor containing binary values that act as indicator variables
        Contains Indicator variable such that I_{u,m_i} = 1 if(u,mi) ∈ E'_{m_i} and I{u,m_i} = 0 otherwise.
        - mst_data :
        Labels indicating the correct and negative candidates for the given contexts. Used for computing loss during training.
        - pos_neg_loss : bool
        Flag to indicate whether to use a specific loss function that considers both positive and negative samples distinctly.
        - only_digits : bool
        Flag to return only the logits (scores) without computing the loss. Useful for evaluation or inference.
        """
        if mst_data is not None:
            context_embeds = self.encode_context(
                context_input, requires_grad=True
            ).unsqueeze(
                2
            )  # batchsize x embed_size x 1

            pos_embeds = mst_data["positive_embeds"].unsqueeze(
                1
            )  # batchsize x 1 x embed_size # this is POSITIVE EMBEDDING !

            neg_dict_embeds = self.encode_candidate(
                mst_data["negative_dict_inputs"], requires_grad=True
            )  # (batchsize*knn_dict) x embed_size : need reshaping

            neg_dict_embeds = neg_dict_embeds.view(
                context_embeds.shape[0],
                neg_dict_embeds.shape[0] // context_embeds.shape[0],
                neg_dict_embeds.shape[1],
            )  # batchsize x knn_dict x embed_size

            cand_embeds = torch.cat(
                (pos_embeds, neg_dict_embeds), dim=1
            )  # batchsize x knn_dict+1 x embed_size

            if mst_data["negative_men_inputs"] is not None:
                neg_men_embeds = self.encode_context(
                    mst_data["negative_men_inputs"], requires_grad=True
                )  # (batchsize*knn_men) x embed_size

                neg_men_embeds = neg_men_embeds.view(
                    context_embeds.shape[0],
                    neg_men_embeds.shape[0] // context_embeds.shape[0],
                    neg_men_embeds.shape[1],
                )  # batchsize x knn_men x embed_size

                cand_embeds = torch.cat(
                    (cand_embeds, neg_men_embeds), dim=1
                )  # batchsize x (knn_men + knn_ent + 1) x embed_size #DD12

            # Compute scores
            scores = torch.bmm(cand_embeds, context_embeds)  # batchsize x topk x 1
            scores = torch.squeeze(scores, dim=2)  # batchsize x topk

        else:  # Case 2
            flag = label_input is None  # True means no negative examples
            scores = self.score_candidate(context_input, cand_input, flag)
            bs = scores.size(0)  # batch_size

        if only_logits:
            return scores

        if label_input is None:
            target = torch.LongTensor(torch.arange(bs))
            ###### The target is np.array([0,1, ..., bs-1]) which is clearly not the real target.
            ###### This if is probably never used as we are interested in the mst performance.
            loss = F.cross_entropy(
                scores, target, reduction="mean"
            )  # F = torch.nn.functional
        else:
            if pos_neg_loss:  # The one always used.
                loss = torch.mean(
                    torch.sum(
                        -torch.log(torch.softmax(scores, dim=1) + 1e-8) * label_input
                        - torch.log(1 - torch.softmax(scores, dim=1) + 1e-8)
                        * (1 - label_input),
                        dim=1,
                    )
                )
            else:  # Uses a simpler loss calculation focusing only on the negative log likelihood of the positive labels
                # Suitable for more traditional classification tasks where the focus is on maximizing the probability of the correct class.
                loss = torch.mean(
                    torch.max(
                        -torch.log(torch.softmax(scores, dim=1) + 1e-8) * label_input,
                        dim=1,
                    )[0]
                )
        return loss, scores


def to_bert_input(token_idx, null_idx):
    """
    Prepare token_idx, segment_idx and mask for encoding
    ------
    Params :
    - token_idx : Tensor of dim (batch_size, sequence_length)
    Contains INDICES of input tokens.
    - segment_idx : Tensor of dim (batch_size, sequence_length)
    For question answering : A / B
    - null_idx : int
    Represents the index value used to indicate padding.
    - mask : Tensor of bool
    Indicates which tokens should be attended to (True) and which should be ignored (False) during the processing by BERT.
    """
    segment_idx = token_idx * 0
    mask = token_idx != null_idx
    token_idx = token_idx * mask.long()
    return token_idx, segment_idx, mask



from bioel.models.arboel.model.common.params import BlinkParser

import lightning as L
from pytorch_lightning.callbacks import ModelCheckpoint

from bioel.models.arboel.crossencoder.original.my_train_cross import (
    CrossEncoderDataModule,
    CrossEncoder,
)


def main(args):
    data_module = CrossEncoderDataModule(params=args)

    MyModel = CrossEncoder.load_from_checkpoint(
        params=args, checkpoint_path=args["model_checkpoint"]
    )

    trainer = L.Trainer(
        # limit_test_batches=4,  # for ncbi_disease
        devices=args["devices"],
        accelerator="gpu",
        strategy="ddp_find_unused_parameters_true",
        enable_progress_bar=True,
        precision="16-mixed",
    )

    trainer.test(model=MyModel, datamodule=data_module)


if __name__ == "__main__":
    parser = BlinkParser(add_model_args=True)
    parser.add_eval_args()
    args = parser.parse_args()
    print(args)
    main(args.__dict__)
